import numpy as np

eps = 1e-10

def properties(chord, tw_aero_d, tw_prime_d, le_loc, xnode, ynode, e1, e2, g12, anu12, density, xsec_nodeU, n_laminaU, n_pliesU, t_lamU, tht_lamU, mat_lamU, xsec_nodeL, n_laminaL, n_pliesL, t_lamL, tht_lamL, mat_lamL, nweb, loc_web, n_laminaW, n_pliesW, t_lamW, tht_lamW, mat_lamW):

    chord = np.array(chord)
    tw_aero_d = np.array(tw_aero_d)
    tw_prime_d = np.array(tw_prime_d)
    le_loc = np.array(le_loc)
    xnode = np.array(xnode)
    ynode = np.array(ynode)
    e1 = np.array(e1)
    e2 = np.array(e2)
    g12 = np.array(g12)
    anu12 = np.array(anu12)
    density = np.array(density)
    xsec_nodeU = np.array(xsec_nodeU)
    t_lamU = np.array(t_lamU)
    tht_lamU = np.array(tht_lamU)
    xsec_nodeL = np.array(xsec_nodeL)
    t_lamL = np.array(t_lamL)
    tht_lamL = np.array(tht_lamL)
    loc_web = np.array(loc_web)
    t_lamW = np.array(t_lamW)
    tht_lamW = np.array(tht_lamW)
    
    n_laminaU = np.array(n_laminaU, dtype=np.int_)
    n_pliesU = np.array(n_pliesU, dtype=np.int_)
    mat_lamU = np.array(mat_lamU, dtype=np.int_)
    n_laminaL = np.array(n_laminaL, dtype=np.int_)
    n_pliesL = np.array(n_pliesL, dtype=np.int_)
    mat_lamL = np.array(mat_lamL, dtype=np.int_)
    n_laminaW = np.array(n_laminaW, dtype=np.int_)
    n_pliesW = np.array(n_pliesW, dtype=np.int_)
    mat_lamW = np.array(mat_lamW, dtype=np.int_)

    max_sectors = np.int_(np.max([n_laminaU.size, n_laminaL.size, n_laminaW.size]))
    max_laminates = np.int_(np.max([n_laminaU.max(), n_laminaL.max(), n_laminaW.max()]))

    n_materials = len(e1)
    n_af_nodes = len(xnode)
    n_sctU = len(n_laminaU)
    n_sctL = len(n_laminaL)
    
    r2d = 180.0 / np.pi

    webs_exist = nweb > 0

    if le_loc < 0.0:
        print(' WARNING** leading edge aft of reference axis **')

    if np.any(anu12 > np.sqrt(e1/e2)):
        idx = np.where(anu12 > np.sqrt(e1/e2))[0]
        raise ValueError(f'**ERROR** material {idx+1} properties not consistent')

    if n_af_nodes <= 2:
        raise ValueError(' ERROR** min 3 nodes reqd to define airfoil geom')

    location = np.argmin(xnode)
    if location != 0:
        raise ValueError(' ERROR** the first airfoil node not a leading node')

    if np.abs(xnode[0]) > eps or np.abs(ynode[0]) > eps:
        raise ValueError(' ERROR** leading-edge node not located at (0,0)')

    location = np.argmax(xnode)
    if xnode.max() > 1.0:
        raise ValueError(' ERROR** trailing-edge node exceeds chord boundary')

    tenode_u = location
    ncounter = np.where(np.abs(xnode[location:] - xnode[location]) < eps)[0].max()

    tenode_l = tenode_u + ncounter
    nodes_u = xnode[:(tenode_u+1)].size
    nodes_l = xnode[tenode_l:].size

    xnode_u = xnode[:nodes_u]
    ynode_u = ynode[:nodes_u]
    xnode_l = np.r_[xnode[0], np.flipud(xnode[tenode_l:])]
    ynode_l = np.r_[ynode[0], np.flipud(ynode[tenode_l:])]

    if np.any(np.abs(np.diff(xnode_u) <= eps)):
        raise ValueError(' ERROR** upper surface not single-valued')

    if np.any(np.abs(np.diff(xnode_l) <= eps)):
        raise ValueError(' ERROR** lower surface not single-valued')

    if ynode_u[1]/xnode_u[1] <= ynode_l[1]/xnode_l[1]:
        raise ValueError(' ERROR** airfoil node numbering not clockwise')

    yinterp_lu = np.interp(xnode_l, xnode_u, ynode_u)
    if np.any(ynode_l[1:-1] >= yinterp_lu[1:-1]):
        raise ValueError(' ERROR** airfoil shape self-crossing')

    if webs_exist:
        # Vectorize the embed
        weby_u = np.interp(loc_web, xnode_u, ynode_u)
        weby_l = np.interp(loc_web, xnode_l, ynode_l)
        xnode_u, idx = np.unique(np.r_[xnode_u, loc_web], return_index=True)
        ynode_u = (np.r_[ynode_u, weby_u])[idx]
        xnode_l, idx = np.unique(np.r_[xnode_l, loc_web], return_index=True)
        ynode_l = (np.r_[ynode_l, weby_l])[idx]
    else:
        weby_u = weby_l = np.array([])

    n_scts = np.array([n_sctU, n_sctL])
    xsec_node = np.zeros((2, max_sectors+1))
    xsec_node[0,:xsec_nodeU.size] = xsec_nodeU
    xsec_node[1,:xsec_nodeL.size] = xsec_nodeL
    if np.any(n_scts <= 0):
        raise ValueError(' ERROR** no of sectors not positive')
    if np.any(xsec_node[:, 0] < 0.0):
        raise ValueError(' ERROR** sector node x-location not positive')
    if np.any(np.diff(xsec_nodeU) <= 0.0):
        raise ValueError(' ERROR** upper sector nodal x-locations not in ascending order')
    if np.any(np.diff(xsec_nodeL) <= 0.0):
        raise ValueError(' ERROR** lower sector nodal x-locations not in ascending order')

    n_weblams = n_laminaW    
    n_laminas = np.zeros((2, max_sectors), dtype=np.int_)
    n_laminas[0,:n_laminaU.size] = n_laminaU
    n_laminas[1,:n_laminaL.size] = n_laminaL
    
    tht_lam = np.zeros((2, max_sectors, max_laminates))
    tlam = np.zeros((2, max_sectors, max_laminates))
    mat_id = np.zeros((2, max_sectors, max_laminates), dtype=np.int_)
    wmat_id = np.zeros((nweb, 6), dtype=np.int_)
    tht_wlam = np.zeros((nweb, 6))
    twlam = np.zeros((nweb, 6))

    k = 0
    for i in range(n_sctU):
        for j in range(n_laminaU[i]):
            tlam[0, i, j] = n_pliesU[k] * t_lamU[k]
            tht_lam[0, i, j] = tht_lamU[k] / r2d
            mat_id[0, i, j] = mat_lamU[k] - 1 # Input is 1-based indexing for Fortran
            k += 1

    k = 0
    for i in range(n_sctL):
        for j in range(n_laminaL[i]):
            tlam[1, i, j] = n_pliesL[k] * t_lamL[k]
            tht_lam[1, i, j] = tht_lamL[k] / r2d
            mat_id[1, i, j] = mat_lamL[k] - 1 # Input is 1-based indexing for Fortran
            k += 1

    k = 0
    for i in range(nweb):
        for j in range(n_laminaW[i]):
            twlam[i, j] = n_pliesW[k] * t_lamW[k]
            tht_wlam[i, j] = tht_lamW[k] / r2d
            wmat_id[i, j] = mat_lamW[k] - 1 # Input is 1-based indexing for Fortran
            k += 1

    xu1 = xsec_node[0, 0]
    xu2 = xsec_node[0, n_sctU]
    if xu2 > xnode_u[-1]:
        raise ValueError(f' ERROR** upper-surf last sector node out of bounds {xu2} {xnode_u[-1]}')
    xl1 = xsec_node[1, 0]
    xl2 = xsec_node[1, n_sctL]
    if xl2 > xnode_l[-1]:
        raise ValueError(f' ERROR** lower-surf last sector node out of bounds {xl2} {xnode_l[-1]}')
            
    # Vectorize the embed
    yinterp_u = np.interp(xsec_nodeU, xnode_u, ynode_u)
    xnode_u, idx = np.unique(np.r_[xnode_u, xsec_nodeU], return_index=True)
    ynode_u = (np.r_[ynode_u, yinterp_u])[idx]
    yu1 = yinterp_u[0]
    yu2 = yinterp_u[-1]
    ndu1 = np.searchsorted(xnode_u, xu1)
    ndu2 = np.searchsorted(xnode_u, xu2)

    yinterp_l = np.interp(xsec_nodeL, xnode_l, ynode_l)
    xnode_l, idx = np.unique(np.r_[xnode_l, xsec_nodeL], return_index=True)
    ynode_l = (np.r_[ynode_l, yinterp_l])[idx]
    yl1 = yinterp_l[0]
    yl2 = yinterp_l[-1]
    ndl1 = np.searchsorted(xnode_l, xl1)
    ndl2 = np.searchsorted(xnode_l, xl2)

    nseg_u = ndu2 - ndu1
    nseg_l = ndl2 - ndl1
    nseg_p = nseg_u + nseg_l
    nseg = nseg_p + nweb if webs_exist else nseg_p
    
    if np.abs(xu1 - xl1) > eps:
        print(' WARNING** the leading edge may be open; check closure')
    else:
        if (yu1 - yl1) > eps:
            wreq = 1
            if webs_exist:
                if np.abs(xu1 - loc_web[0]) < eps:
                    wreq = 0
            if wreq == 1:
                print(' WARNING** open leading edge; check web requirement')

    if np.abs(xu2 - xl2) > eps:
        print(' WARNING** the trailing edge may be open; check closure')
    else:
        if (yu2 - yl2) > eps:
            wreq = 1
            if webs_exist:
                if np.abs(xu2 - loc_web[-1]) < eps:
                    wreq = 0
            if wreq == 1:
                print(' WARNING** open trailing edge; check web requirement')

    if webs_exist:
        if loc_web[0] < xu1 or loc_web[0] < xl1:
            print('ERROR** first web out of sectors-bounded airfoil')
        if loc_web[-1] > xu2 or loc_web[-1] > xl2:
            print(' ERROR** last web out of sectors-bounded airfoil')

    # Begin segment calculations
    tw_aero = tw_aero_d / r2d
    tw_prime = tw_prime_d / r2d
    tphip = tw_prime

    anud = 1.0 - anu12 * anu12 * e2 / e1
    q11 = e1 / anud
    q22 = e2 / anud
    q12 = anu12 * e2 / anud
    q66 = g12

    isur, idsect, yseg, zseg, wseg, sthseg, cthseg, s2thseg, c2thseg = seg_info(chord, le_loc, nseg, nseg_u, nseg_p, xnode_u, ynode_u, xnode_l, ynode_l, ndl1, ndu1, loc_web, weby_u, weby_l, n_scts, xsec_node)

    # Initialize variables
    sigma = 0.0
    eabar = 0.0
    q11ya = 0.0
    q11za = 0.0
    
    #   segments loop for sc
    for iseg in range(nseg_p):
        # Initialize segment properties
        ks = isur[iseg]
        idsec = idsect[iseg]
        ysg = yseg[iseg]
        zsg = zseg[iseg]
        w = wseg[iseg]
        sths = sthseg[iseg]
        cths = cthseg[iseg]
        nlam = n_laminas[ks, idsec]

        # initialization for seg (sc)
        tbar = 0.0
        q11t = 0.0
        q11yt_u = 0.0
        q11zt_u = 0.0
        q11yt_l = 0.0
        q11zt_l = 0.0
        
        # Vectorized
        t = tlam[ks, idsec, :]       # thickness
        thp = tht_lam[ks, idsec, :]  # ply angle
        mat = mat_id[ks, idsec, :]   # material no.
        tbar = np.cumsum(np.r_[0, t[:-1]]) + (t / 2.0) # Just need half of the current increment
        y0 = ysg - ((-1.0) ** (ks+1)) * tbar * sths
        z0 = zsg + ((-1.0) ** (ks+1)) * tbar * cths
        
        qbar11, qbar22, qbar12, qbar16, qbar26, qbar66 = q_bars(thp, q11[mat], q22[mat], q12[mat], q66[mat])
        qtil = q_tildas(qbar11, qbar22, qbar12, qbar16, qbar26, qbar66)

        qtil11t = np.squeeze(qtil[0, 0, :] * t)
        q11t = np.sum(qtil11t)

        if iseg < nseg_u:
            q11yt_u = np.sum(qtil11t * y0)
            q11zt_u = np.sum(qtil11t * z0)
        else:
            q11yt_l = np.sum(qtil11t * y0)
            q11zt_l = np.sum(qtil11t * z0)

        tbar = t.sum() # Now need the full thing
        '''
        # Original python port with inner loop
        for ilam in range(nlam):
            t = tlam[ks, idsec, ilam]       # thickness
            thp = tht_lam[ks, idsec, ilam]  # ply angle
            mat = mat_id[ks, idsec, ilam]   # material no.

            tbar = tbar + t / 2.0
            y0 = ysg - ((-1.0) ** (ks+1)) * tbar * sths
            z0 = zsg + ((-1.0) ** (ks+1)) * tbar * cths

            # Call function q_bars and q_tildas
            qbar11, qbar22, qbar12, qbar16, qbar26, qbar66 = q_bars(thp, q11[mat], q22[mat], q12[mat], q66[mat])
            qtil = q_tildas(qbar11, qbar22, qbar12, qbar16, qbar26, qbar66)

            qtil11t = qtil[0, 0] * t
            q11t = q11t + qtil11t

            if iseg < nseg_u:
                q11yt_u = q11yt_u + qtil11t * y0
                q11zt_u = q11zt_u + qtil11t * z0
            else:
                q11yt_l = q11yt_l + qtil11t * y0
                q11zt_l = q11zt_l + qtil11t * z0

            tbar = tbar + t / 2.0
        '''

        # add seg contributions (sc)
        sigma = sigma + w * np.abs(zsg + ((-1.0) ** (ks+1)) * 0.5 * tbar * cths) * cths
        eabar = eabar + q11t * w
        q11ya = q11ya + (q11yt_u + q11yt_l) * w
        q11za = q11za + (q11zt_u + q11zt_l) * w

    # get section sc
    y_sc = q11ya / eabar
    z_sc = q11za / eabar
    #---------------- end section sc -----------

    #   initializations for section (properties)
    eabar = 0.0
    q11ya = 0.0
    q11za = 0.0
    ap = 0.0
    bp = 0.0
    cp = 0.0
    dp = 0.0
    ep = 0.0
    q11ysqa = 0.0
    q11zsqa = 0.0
    q11yza = 0.0
    mass = 0.0
    area = 0.0
    rhoya = 0.0
    rhoza = 0.0
    rhoysqa = 0.0
    rhozsqa = 0.0
    rhoyza = 0.0
    
    #   segments loop (for properties)
    for iseg in range(nseg):
        # Initialize segment properties
        ks = isur[iseg]
        idsec = idsect[iseg]
        ysg = yseg[iseg]
        zsg = zseg[iseg]
        w = wseg[iseg]
        sths = sthseg[iseg]
        cths = cthseg[iseg]
        s2ths = s2thseg[iseg]
        c2ths = c2thseg[iseg]

        nlam = n_laminas[ks, idsec] if ks >= 0 else n_weblams[idsec]

        tbar = 0.0
        q11t = 0.0
        q11yt = 0.0
        q11zt = 0.0
        dtbar = 0.0
        q2bar = 0.0
        zbart = 0.0
        ybart = 0.0
        tbart = 0.0
        q11ysqt = 0.0
        q11zsqt = 0.0
        q11yzt = 0.0
        rhot = 0.0
        rhoyt = 0.0
        rhozt = 0.0
        rhoysqt = 0.0
        rhozsqt = 0.0
        rhoyzt = 0.0

        # Vectorized
        if ks >= 0:
            t = tlam[ks, idsec, :]       # thickness
            thp = tht_lam[ks, idsec, :]  # ply angle
            mat = mat_id[ks, idsec, :]   # material no.
            tbar = np.cumsum(np.r_[0, t[:-1]]) + (t / 2.0) # Just need half of the current increment
            y0 = ysg - ((-1.0) ** (ks+1)) * tbar * sths - y_sc
            z0 = zsg + ((-1.0) ** (ks+1)) * tbar * cths - z_sc
        else:
            t = twlam[idsec, :]
            thp = tht_wlam[idsec, :]
            mat = wmat_id[idsec, :]
            tbar = np.cumsum(np.r_[0, t[:-1]]) + (t / 2.0) # Just need half of the current increment
            y0 = ysg - tbar / 2.0 - y_sc
            z0 = zsg - z_sc

        y0sq = y0 * y0
        z0sq = z0 * z0

        qbar11, qbar22, qbar12, qbar16, qbar26, qbar66 = q_bars(thp, q11[mat], q22[mat], q12[mat], q66[mat])
        qtil = q_tildas(qbar11, qbar22, qbar12, qbar16, qbar26, qbar66)

        ieta1 = (t ** 2) / 12.0
        izeta1 = (w ** 2) / 12.0
        iepz = 0.5 * (ieta1 + izeta1)
        iemz = 0.5 * (ieta1 - izeta1)
        ipp = iepz + iemz * c2ths
        iqq = iepz - iemz * c2ths
        ipq = iemz * s2ths
        qtil11t = np.squeeze(qtil[0, 0, :] * t)
        rot = density[mat] * t

        if ks >= 0:
            qtil12t = np.squeeze(qtil[0, 1, :]) * t
            qtil22t = np.squeeze(qtil[1, 1, :]) * t

            q11t = np.sum(qtil11t)
            q11yt = np.sum(qtil11t * y0)
            q11zt = np.sum(qtil11t * z0)
            dtbar = np.sum(qtil12t * (y0sq + z0sq) * tphip * t)
            q2bar = np.sum(qtil22t)
            zbart = np.sum(z0 * qtil12t)
            ybart = np.sum(y0 * qtil12t)
            tbart = np.sum(qtil12t)
            q11ysqt = np.sum(qtil11t * (y0sq + iqq))
            q11zsqt = np.sum(qtil11t * (z0sq + ipp))
            q11yzt = np.sum(qtil11t * (y0 * z0 + ipq))
            rhot = np.sum(rot)
            rhoyt = np.sum(rot * y0)
            rhozt = np.sum(rot * z0)
            rhoysqt = np.sum(rot * (y0sq + iqq))
            rhozsqt = np.sum(rot * (z0sq + ipp))
            rhoyzt = np.sum(rot * (y0 * z0 + ipq))
        else:
            q11t = np.sum(qtil11t)
            q11yt = np.sum(qtil11t * y0)
            q11zt = np.sum(qtil11t * z0)
            q11ysqt = np.sum(qtil11t * (y0sq + iqq))
            q11zsqt = np.sum(qtil11t * (z0sq + ipp))
            q11yzt = np.sum(qtil11t * (y0 * z0 + ipq))
            rhot = np.sum(rot)
            rhoyt = np.sum(rot * y0)
            rhozt = np.sum(rot * z0)
            rhoysqt = np.sum(rot * (y0sq + iqq))
            rhozsqt = np.sum(rot * (z0sq + ipp))
            rhoyzt = np.sum(rot * (y0 * z0 + ipq))
            
        tbar = t.sum() # Now need the full thing

        # Loop over laminates
        '''
        for ilam in range(nlam):
            if ks >= 0:
                t = tlam[ks, idsec, ilam]
                thp = tht_lam[ks, idsec, ilam]
                mat = mat_id[ks, idsec, ilam]
                tbar = tbar + t / 2.0
                y0 = ysg - ((-1.0) ** (ks+1)) * tbar * sths - y_sc
                z0 = zsg + ((-1.0) ** (ks+1)) * tbar * cths - z_sc
            else:
                t = twlam[idsec, ilam]
                thp = tht_wlam[idsec, ilam]
                mat = wmat_id[idsec, ilam]
                tbar = tbar + t / 2.0
                y0 = ysg - tbar / 2.0 - y_sc
                z0 = zsg - z_sc

            y0sq = y0 * y0
            z0sq = z0 * z0

            # Call function q_bars and q_tildas
            qbar11, qbar22, qbar12, qbar16, qbar26, qbar66 = q_bars(thp, q11[mat], q22[mat], q12[mat], q66[mat])
            qtil = q_tildas(qbar11, qbar22, qbar12, qbar16, qbar26, qbar66)

            ieta1 = (t ** 2) / 12.0
            izeta1 = (w ** 2) / 12.0
            iepz = 0.5 * (ieta1 + izeta1)
            iemz = 0.5 * (ieta1 - izeta1)
            ipp = iepz + iemz * c2ths
            iqq = iepz - iemz * c2ths
            ipq = iemz * s2ths
            qtil11t = qtil[0, 0] * t
            rot = density[mat] * t

            if ks >= 0:
                qtil12t = qtil[0, 1] * t
                qtil22t = qtil[1, 1] * t

                q11t = q11t + qtil11t
                q11yt = q11yt + qtil11t * y0
                q11zt = q11zt + qtil11t * z0
                dtbar = dtbar + qtil12t * (y0sq + z0sq) * tphip * t
                q2bar = q2bar + qtil22t
                zbart = zbart + z0 * qtil12t
                ybart = ybart + y0 * qtil12t
                tbart = tbart + qtil12t
                q11ysqt = q11ysqt + qtil11t * (y0sq + iqq)
                q11zsqt = q11zsqt + qtil11t * (z0sq + ipp)
                q11yzt = q11yzt + qtil11t * (y0 * z0 + ipq)
                rhot = rhot + rot
                rhoyt = rhoyt + rot * y0
                rhozt = rhozt + rot * z0
                rhoysqt = rhoysqt + rot * (y0sq + iqq)
                rhozsqt = rhozsqt + rot * (z0sq + ipp)
                rhoyzt = rhoyzt + rot * (y0 * z0 + ipq)
            else:
                q11t = q11t + qtil11t
                q11yt = q11yt + qtil11t * y0
                q11zt = q11zt + qtil11t * z0
                q11ysqt = q11ysqt + qtil11t * (y0sq + iqq)
                q11zsqt = q11zsqt + qtil11t * (z0sq + ipp)
                q11yzt = q11yzt + qtil11t * (y0 * z0 + ipq)
                rhot = rhot + rot
                rhoyt = rhoyt + rot * y0
                rhozt = rhozt + rot * z0
                rhoysqt = rhoysqt + rot * (y0sq + iqq)
                rhozsqt = rhozsqt + rot * (z0sq + ipp)
                rhoyzt = rhoyzt + rot * (y0 * z0 + ipq)

            tbar = tbar + t / 2.0
        '''

        eabar = eabar + q11t * w
        q11ya = q11ya + q11yt * w
        q11za = q11za + q11zt * w
        q11ysqa = q11ysqa + q11ysqt * w
        q11zsqa = q11zsqa + q11zsqt * w
        q11yza = q11yza + q11yzt * w

        if ks >= 0:
            wdq2bar = w / q2bar
            ap = ap + wdq2bar
            bp = bp + wdq2bar * tbart
            cp = cp + wdq2bar * dtbar
            dp = dp + wdq2bar * zbart
            ep = ep + wdq2bar * ybart

        area = area + w
        mass = mass + rhot * w
        rhoya = rhoya + rhoyt * w
        rhoza = rhoza + rhozt * w
        rhoysqa = rhoysqa + rhoysqt * w
        rhozsqa = rhozsqa + rhozsqt * w
        rhoyza = rhoyza + rhoyzt * w

    y_tc = q11ya / eabar
    z_tc = q11za / eabar
    
    sfbar = q11za
    slbar = q11ya
    eifbar = q11zsqa
    eilbar = q11ysqa
    eiflbar = q11yza
    
    sigm2 = sigma*2.0
    gjbar = sigm2*(sigm2+cp)/ap
    sftbar = -sigm2*dp/ap
    sltbar = -sigm2*ep/ap
    satbar = sigm2*bp/ap
    
    ycm_sc =   rhoya/mass #wrt sc
    zcm_sc =   rhoza/mass #wrt sc
    
    iflap_sc = rhozsqa #wrt sc
    ilag_sc = rhoysqa   #wrt sc
    ifl_sc = rhoyza     #wrt sc
    
    # get section tc and cm
    
    ytc_ref =  y_tc + y_sc  #wrt the ref axes
    ztc_ref =  z_tc + z_sc  #wrt the ref axes
    
    ycm_ref =  ycm_sc + y_sc    #wrt the ref axes
    zcm_ref =  zcm_sc + z_sc    #wrt the ref axes
    
    # moments of inertia # about ref_parallel axes at cm
    
    iflap_cm = iflap_sc - mass*zcm_sc**2
    ilag_cm = ilag_sc - mass*ycm_sc**2
    ifl_cm = ifl_sc - mass*ycm_sc*zcm_sc
    
    # inertia principal axes orientation and moments of inertia
    
    m_inertia = 0.5*(ilag_cm + iflap_cm)
    r_inertia = np.sqrt(0.25*((ilag_cm-iflap_cm)**2) + ifl_cm**2)
    
    if (iflap_cm <= ilag_cm):
        iflap_eta = m_inertia - r_inertia
        ilag_zeta = m_inertia + r_inertia
    else:
        iflap_eta = m_inertia + r_inertia
        ilag_zeta = m_inertia - r_inertia

    if (ilag_cm == iflap_cm):
        th_pa = np.pi/4.0
        if (np.abs(ifl_cm/iflap_cm) < 1e-6):
            th_pa = 0.0
    else:
        th_pa = 0.5*np.abs(np.arctan(2.0*ifl_cm/(ilag_cm-iflap_cm)))

    if (np.abs(ifl_cm) < eps):
        th_pa = 0.0
    else:
        if (iflap_cm >= ilag_cm and ifl_cm > 0.):
            th_pa = -th_pa
        if (iflap_cm >= ilag_cm and ifl_cm < 0.):
            th_pa = th_pa
        if (iflap_cm < ilag_cm and ifl_cm > 0.):
            th_pa = th_pa
        if (iflap_cm < ilag_cm and ifl_cm < 0.):
            th_pa = -th_pa

    # elastic principal axes orientation and principal bending stiffneses

    em_stiff = 0.5*(eilbar + eifbar)
    er_stiff = np.sqrt(0.25*((eilbar-eifbar)**2) + eiflbar**2)
    
    if (eifbar <= eilbar):
        pflap_stff = em_stiff - er_stiff
        plag_stff = em_stiff + er_stiff
    else:
        pflap_stff = em_stiff + er_stiff
        plag_stff = em_stiff - er_stiff

    if (eilbar == eifbar):
        the_pa = np.pi/4.0
    else:
        the_pa = 0.5*np.abs(np.arctan(2.0*eiflbar/(eilbar-eifbar)))

    if (np.abs(eiflbar) < eps):
        the_pa = 0.0
    else:
        if (eifbar >= eilbar and eiflbar > 0.):
            the_pa = -the_pa
        if (eifbar >= eilbar and eiflbar < 0.):
            the_pa = the_pa
        if (eifbar < eilbar and eiflbar > 0.):
            the_pa = the_pa
        if (eifbar < eilbar and eiflbar < 0.):
            the_pa = -the_pa

    #---------------- end properties computation -----------

    # ---------- prepare outputs --------------
    id_form = 1  # hardwired for wt's

    if (id_form == 1):
        tw_iner = tw_aero - th_pa
        str_tw =  tw_aero - the_pa
        y_sc = -y_sc
        ytc_ref = -ytc_ref
        ycm_ref = -ycm_ref
    else:         # for h/c
        #       note: for h/c, th_aero input is +ve acc to h/c convention
        tw_iner = tw_aero + th_pa
        str_tw =  tw_aero + the_pa

    # conversions
    eiflbar = -eiflbar
    sfbar = -sfbar
    sltbar = -sltbar
    tw_iner = tw_iner*r2d

    return (eifbar, eilbar, gjbar, eabar, eiflbar, sfbar, slbar, sftbar, sltbar, satbar, z_sc, y_sc, ztc_ref, ytc_ref, mass, area, iflap_eta, ilag_zeta, tw_iner, zcm_ref, ycm_ref)

def seg_info(ch, rle, nseg, nseg_u, nseg_p, xnode_u, ynode_u, xnode_l, ynode_l, ndl1, ndu1, loc_web, weby_u, weby_l, n_scts, xsec_node):
    # NOTE: coord transformation from xaf-yaf to yre-zref and seg info

    # inputs
    # real(dbp), intent(in) :: ch, rle  # chord length, loc of l.e. (non-d wrt chord)
    # integer, intent(in) :: nseg, nseg_u, nseg_p  # total number of segs, no of segs on the upper surface, no of segs for both upper and lower surfaces
    # real(dbp), intent(in), dimension(300) :: xnode_u, ynode_u, xnode_l, ynode_l  # x,y nodes on upper/lower
    # integer, intent(in) :: ndl1, ndu1 # 1st seg lhs node number lower/upper surface
    # real(dbp), intent(in), dimension(:) :: loc_web, weby_u, weby_l  # x coord of web, y coord of web upper/lower
    # integer, intent(in), dimension(2) :: n_scts  # no of sectors on 'is' surf
    # integer, intent(in) :: nsecnode
    # real(dbp), dimension(2, nsecnode) :: xsec_node  # x coord of sect-i lhs on 's' surf

    # outputs
    '''
    isur = -1 * np.ones(nseg, dtype=np.int_)  # surf id
    thseg = -np.pi / 2.0 * np.ones(nseg)

    # Vectorized
    iseg = np.arange(nseg, dtype=np.int_)
    isur[iseg < nseg_p] = 1 # Do this first
    isur[iseg < nseg_u] = 0
    nd_a = np.zeros(nseg, dtype=np.int_)
    nd_a[iseg < nseg_p] = ndl1 + iseg[iseg < nseg_p] - nseg_u
    nd_a[iseg < nseg_u] = ndu1 + iseg[iseg < nseg_u]
    xa = xnode_u[nd_a]
    ya = ynode_u[nd_a]
    xb = xnode_u[nd_a + 1]
    yb = ynode_u[nd_a + 1]
    
    iweb = iseg[iseg >= nseg_p] - nseg_p
    xa[iseg >= nseg_p] = loc_web[iweb]
    xb[iseg >= nseg_p] = xa[iseg >= nseg_p]
    ya[iseg >= nseg_p] = weby_u[iweb]
    yb[iseg >= nseg_p] = weby_l[iweb]

    xba = xb - xa
    yba = ya - yb
    yseg = ch * (2. * rle - xa - xb) / 2.0  # yref coord of mid-seg pt (in r-frame)
    zseg = ch * (ya + yb) / 2.0  # zref coord of mid-seg pt (in r-frame)
    wseg = ch * np.sqrt(xba ** 2 + yba ** 2)

    thseg[isur >= 0] = np.arctan(yba[isur >= 0] / xba[isur >= 0])  # thseg +ve in new y-z ref frame
    
    idsect = np.zeros(nseg, dtype=np.int_)  # associated sect or web number
    idsect[iseg >= nseg_p] = iweb
    for iseg in range(nseg_p):  # seg numbering from le clockwise
        ks = isur[iseg]
        for i in range(n_scts[ks]):
            if xa[iseg] > (xsec_node[ks, i] - eps) and xb[iseg] < (xsec_node[ks, i + 1] + eps):
                idsect[iseg] = i
                break

    '''
    isur = np.zeros(nseg, dtype=np.int_)  # surf id
    thseg = np.zeros(nseg)
    idsect = np.zeros(nseg, dtype=np.int_)  # associated sect or web number
    yseg = np.zeros(nseg)  # y-ref of mid-seg point
    zseg = np.zeros(nseg)  # z-ref of mid-seg point
    wseg = np.zeros(nseg)  # seg width
    
    for iseg in range(nseg):  # seg numbering from le clockwise
        ks = -2
        if iseg < nseg_u:  # upper surface segs
            nd_a = ndu1 + iseg
            xa = xnode_u[nd_a]
            ya = ynode_u[nd_a]
            xb = xnode_u[nd_a + 1]
            yb = ynode_u[nd_a + 1]
            ks = 0
        elif iseg < nseg_p:  # lower surface segs
            nd_a = ndl1 + iseg - nseg_u
            xa = xnode_l[nd_a]  # xref of node toward le (in a/f ref frame)
            ya = ynode_l[nd_a]  # yref of node toward le (in new ref frame)
            xb = xnode_l[nd_a + 1]  # xref of node toward te (in a/f ref frame)
            yb = ynode_l[nd_a + 1]  # yref of node toward te (in new ref frame)
            ks = 1
        else:
            iweb = iseg - nseg_p
            xa = loc_web[iweb]
            xb = xa
            ya = weby_u[iweb]
            yb = weby_l[iweb]
            ks = -1

        if ks == -2:
            print(f'iseg={iseg}')
            raise ValueError('ERROR** unknown, contact NREL')

        isur[iseg] = ks

        if ks >= 0:  # id associated sect number
            icheck = True
            for i in range(n_scts[ks]):
                if xa > (xsec_node[ks, i] - eps) and xb < (xsec_node[ks, i + 1] + eps):
                    idsect[iseg] = i
                    icheck = False
                    break

            if icheck:
                print('ERROR** unknown, contact NREL')
        else:
            idsect[iseg] = iweb  # id associated web number

        xba = xb - xa
        yba = ya - yb
        yseg[iseg] = ch * (2. * rle - xa - xb) / 2.0  # yref coord of mid-seg pt (in r-frame)
        zseg[iseg] = ch * (ya + yb) / 2.0  # zref coord of mid-seg pt (in r-frame)
        wseg[iseg] = ch * np.sqrt(xba ** 2 + yba ** 2)

        if ks >= 0:  # id associated sect number
            thseg[iseg] = np.arctan(yba / xba)  # thseg +ve in new y-z ref frame
        else:
            thseg[iseg] = -np.pi / 2.0

    sthseg = np.sin(thseg)
    cthseg = np.cos(thseg)
    s2thseg = np.sin(2.0 * thseg)
    c2thseg = np.cos(2.0 * thseg)

    return isur, idsect, yseg, zseg, wseg, sthseg, cthseg, s2thseg, c2thseg


def tw_rate(sloc, tw_aero):
    th_prime = np.zeros(sloc.shape)  # vector of twist rates

    f0 = tw_aero[1:-1]
    f1 = tw_aero[0:-2]
    f2 = tw_aero[2:]
    h1 = sloc[1:-1] - sloc[0:-2]
    h2 = sloc[2:] - sloc[1:-1]
    th_prime[1:-1] = (h1*(f2-f0) + h2*(f0-f1))/(2.*h1*h2)

    #for i in range(1, naf-1):
    #    f0 = tw_aero[i]
    #    f1 = tw_aero[i-1]
    #    f2 = tw_aero[i+1]
    #    h1 = sloc[i] - sloc[i-1]
    #    h2 = sloc[i+1] - sloc[i]
    #    th_prime[i] = (h1*(f2-f0) + h2*(f0-f1))/(2.*h1*h2)

    th_prime[0] = (tw_aero[1]-tw_aero[0])/(sloc[1]-sloc[0])
    th_prime[-1] = (tw_aero[-1]-tw_aero[-2])/(sloc[-1]-sloc[-2])

    return th_prime


def q_bars(thp, q11, q22, q12, q66):
    ct = np.cos(thp)
    st = np.sin(thp)

    c2t = ct**2
    c3t = c2t * ct
    c4t = c3t * ct
    s2t = st**2
    s3t = s2t * st
    s4t = s3t * st
    s2thsq = 4.0 * s2t * c2t

    k11 = q11
    k22 = q22
    k12 = q12
    k66 = q66
    kmm = k11 - k12 - 2.0 * k66
    kmp = k12 - k22 + 2.0 * k66

    qbar11 = k11 * c4t + 0.5 * (k12 + 2.0 * k66) * s2thsq + k22 * s4t
    qbar22 = k11 * s4t + 0.5 * (k12 + 2.0 * k66) * s2thsq + k22 * c4t
    qbar12 = 0.25 * (k11 + k22 - 4.0 * k66) * s2thsq + k12 * (s4t + c4t)
    qbar16 = kmm * st * c3t + kmp * s3t * ct
    qbar26 = kmm * s3t * ct + kmp * st * c3t
    qbar66 = 0.25 * (kmm + k22 - k12) * s2thsq + k66 * (s4t + c4t)

    return qbar11, qbar22, qbar12, qbar16, qbar26, qbar66


def q_tildas(qbar11, qbar22, qbar12, qbar16, qbar26, qbar66):
    n = qbar11.size if isinstance(qbar11, type(np.array([]))) else 1        
    qtil = np.zeros((2, 2, n))

    qtil[0, 0, :] = qbar11 - qbar12**2 / qbar22
    qtil[0, 1, :] = qbar16 - qbar12 * qbar26 / qbar22
    qtil[1, 1, :] = qbar66 - qbar26**2 / qbar22
    
    return qtil



"""
Testing framework for the `MooringSystemInstallation` class.
"""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


from copy import deepcopy

import pandas as pd
import pytest

from wisdem.orbit import ProjectManager
from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.core.defaults import process_times as pt
from wisdem.orbit.phases.install import MooringSystemInstallation
from wisdem.test.test_orbit.data import test_weather

config = extract_library_specs("config", "mooring_system_install")


def test_simulation_creation():
    sim = MooringSystemInstallation(config)

    assert sim.config == config
    assert sim.env
    assert sim.port
    assert sim.vessel
    assert sim.num_systems


@pytest.mark.parametrize("weather", (None, test_weather), ids=["no_weather", "test_weather"])
def test_full_run_logging(weather):
    sim = MooringSystemInstallation(config, weather=weather)
    sim.run()

    lines = config["plant"]["num_turbines"] * config["mooring_system"]["num_lines"]

    df = pd.DataFrame(sim.env.actions)
    df = df.assign(shift=(df.time - df.time.shift(1)))
    assert (df.duration - df["shift"]).fillna(0.0).abs().max() < 1e-9
    assert df[df.action == "Install Mooring Line"].shape[0] == lines

    assert ~df["cost"].isnull().any()
    _ = sim.agent_efficiencies
    _ = sim.detailed_output


@pytest.mark.parametrize(
    "anchor, key",
    [
        ("Suction Pile", "suction_pile_install_time"),
        ("Drag Embedment", "drag_embed_install_time"),
    ],
)
def test_kwargs(anchor, key):
    new = deepcopy(config)
    new["mooring_system"]["anchor_type"] = anchor

    sim = MooringSystemInstallation(new)
    sim.run()
    baseline = sim.total_phase_time

    keywords = ["mooring_system_load_time", "mooring_site_survey_time", key]

    failed = []

    for kw in keywords:
        default = pt[kw]
        kwargs = {kw: default + 2}

        new_sim = MooringSystemInstallation(new, **kwargs)
        new_sim.run()
        new_time = new_sim.total_phase_time

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"'{failed}' not affecting results.")

    else:
        assert True


@pytest.mark.parametrize(
    "anchor, key",
    [
        ("Suction Pile", "suction_pile_install_time"),
        ("Drag Embedment", "drag_embed_install_time"),
    ],
)
def test_kwargs_in_ProjectManager(anchor, key):
    base = deepcopy(config)
    base["mooring_system"]["anchor_type"] = anchor
    base["install_phases"] = ["MooringSystemInstallation"]

    project = ProjectManager(base)
    project.run()
    baseline = project.phase_times["MooringSystemInstallation"]

    keywords = ["mooring_system_load_time", "mooring_site_survey_time", key]

    failed = []

    for kw in keywords:
        default = pt[kw]
        processes = {kw: default + 2}
        new_config = deepcopy(base)
        new_config["processes"] = processes

        new_project = ProjectManager(new_config)
        new_project.run()
        new_time = new_project.phase_times["MooringSystemInstallation"]

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"'{failed}' not affecting results.")

    else:
        assert True



import pandas as pd
import numpy as np
from math import ceil

from wisdem.landbosse.model.CostModule import CostModule
from wisdem.landbosse.model.WeatherDelay import WeatherDelay

import traceback

# constants
km_per_m = 0.001
hr_per_min = 1/60
m_per_ft = 0.3048


class Point(object):
    def __init__(self, x, y):
        if type(x) == type(pd.Series(dtype=np.float_)):
            self.x = float(x.values[0])
            self.y = float(y.values[0])
        elif type(x) == type(np.array([])):
            self.x = float(x[0])
            self.y = float(y[0])
        elif type(x) == type(int(0)):
            self.x = float(x)
            self.y = float(y)
        elif type(x) == type(float(0.0)):
            self.x = x
            self.y = y
        else:
            raise ValueError(type(x))

def ccw(A,B,C):
    return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)

# Return true if line segments AB and CD intersect
def intersect(A,B,C,D):
    return ccw(A,C,D) != ccw(B,C,D) and ccw(A,B,C) != ccw(A,B,D)

def point_in_polygon(pt, poly):
    result = False
    maxx = float(np.r_[pt.x, np.array([m.x for m in poly])].max())
    for i in range(len(poly)-1):
        if intersect(poly[i], poly[i+1], pt, Point(1.1*maxx, pt.y)):
            result = not result
    if intersect(poly[-1], poly[0], pt, Point(1.1*maxx, pt.y)):
        result = not result
    return result

class ErectionCost(CostModule):
    """
    ErectionCost.py
    Created by Annika Eberle and Owen Roberts on Mar. 16, 2018
    Created by Alicia Key and Parangat Bhaskar on 01 June 2019

    Calculates the costs for erecting the tower and rotor nacelle assembly for land-based wind projects
    (items in brackets are not yet implemented)

    [Get terrain complexity]
    [Get site complexity]
    Get number of turbines
    Get duration of construction
    Get rate of deliveries
    Get daily hours of operation
    Get turbine rating
    Get component specifications
    [Get crane availability]

    Get price data
        Get labor mobilization_prices by crew type
        Get labor prices by crew type
        Get equipment mobilization prices by equipment type
        Get fuel prices
        Get equipment prices by equipment type

    Calculate operational time for lifting components

    Estimate potential time delays due to weather

    Calculate required labor and equip for erection (see equip_labor_by_type method below)
        Calculate number of workers by crew type
        Calculate man hours by crew type
        Calculate number of equipment by equip type
        Calculate equipment hours by equip type

    Calculate erection costs by type (see methods below)
        Calculate mobilization costs as function of number of workers by crew type, number of equipment by equipment type, labor_mobilization_prices, and equip_mobilization_prices
        Calculate labor costs as function of man_hours and labor prices by crew type
        Calculate fuel costs as function of equipment hours by equipment type and fuel prices by equipment type
        Calculate equipment costs as function of equipment hours by equipment type and equipment prices by equipment type

    Sum erection costs over all types to get total costs

    Find the least cost option

    Return total erection costs

    Keys in the input dictionary are the following:

    construct_duration
        (int) duration of construction (in months)

    rate_of_deliveries
        (int) rate of deliveries (number of turbines per week)

    weather_window
        (pd.DataFrame) window of weather data for project of interest.

    wind_shear_exponent
-        (float) The exponent of the power law wind shear calculation

    overtime_multiplier:
        (float) multiplier for overtime work (working 60 hr/wk vs 40 hr/wk)

    allow_same_flag
        (bool) boolean flag to indicate whether choosing same base and
        topping crane is allowed.

    operational_construction_time
        (int) Number of hours per day when construction can happen.

    time_construct
        (int) 'normal' (10 hours per day) or 'long' (24 hours per day)

    project_data
        (dict) dictionary of pd.DataFrame for each of the csv files loaded
        for the project.

    In turn, the project_data dictionary contains key value pairs of the
    following:

    crane_specs:
        (pd.DateFrame) Specs about the cranes for the cost calculations.

    equip
        (pd.DataFrame) Equipment needed for various tasks

    crew
        (pd.DataFrame) Crew configurations needed for various tasks

    components
        (pd.DataFrame) components to build a wind turbine

    project
        (pd.DataFrame) The project of the project to calculate.

    equip_price
        (pd.DatFrame) Prices to operate various pieces of equipment.

    crew_price
        (pd.DataFrame) THe prices for various crews

    material_price
        (pd.DatFrame) Prices for various materials used during erection.

    rsmeans
        (p.DataFrame) RSMeans data
    """
    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

        # These instance attributes are for diagnostics inside this class
        # only.
        self._possible_crane_cost = None
        self._number_of_equip = None

    def run_module(self):
        """
        Runs the ErectionCost model and populates the IO dictionaries with
        calculated values.

        Returns
        -------
        int
            0 if the module ran successfully. 1 if the module did not run
            successfully
        """
        try:
            self.calculate_costs()
            self.outputs_for_detailed_tab()
            self.output_dict['erection_module_type_operation'] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict['total_erection_cost'],
                project_id=self.project_name,
                total_or_turbine=True
            )
            return 0, 0 # Module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} ErectionCost")
            return 1, error # Module did not run successfully

    def outputs_for_detailed_tab(self):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """
        result =[]

        for _, row in self._number_of_equip.iterrows():
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': '_number_of_equip: Operation-Crane name-Boom system-Number of equipment',
                'value': f'{row["Operation"]}-{row["Crane name"]}-{row["Boom system"]}-{row["Number of equipment"]}',
                'last_number': row["Number of equipment"]
            })

        for _, row in self.output_dict['erection_selected_detailed_data'].iterrows():
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': f'erection_selected_detailed_data: Operation-Crane name-Boom system-Operational construct days over time construct days',
                'value': f'{row["Operation"]}-{row["Crane name"]}-{row["Boom system"]}-{row["Operational construct days over time construct days"]}',
                'last_number': row["Operational construct days over time construct days"]
            })

        for row in self.output_dict['component_name_topvbase'].itertuples():
            dashed_row = '{} - {}'.format(row[1], row[2])
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'component_name_topvbase: Operation - Top or Base',
                'value': dashed_row
            })

        for row in self.output_dict['crane_choice'].itertuples():
            dashed_row = '{} - {} - {}'.format(row[1], row[2], row[3])
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'crane_choice: Crew name - Boom system - Operation',
                'value': dashed_row
            })

        for _, row in self.output_dict['crane_data_output'].iterrows():
            dashed_row = '{} - {} - {}'.format(row.iloc[0], row.iloc[1], row.iloc[2])
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'crane_data_output: crane_boom_operation_concat - variable - value',
                'value': dashed_row,
                'last_number': row.iloc[2]
            })

        for _, row in self.output_dict['crane_cost_details'].iterrows():
            dashed_row = '{} - {} - {}'.format(row.iloc[0], row.iloc[1], row.iloc[2])
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'crane_cost_details: Operation ID - Type of cost - Cost',
                'value': dashed_row,
                'last_number': row.iloc[2]
            })

        for _, row in self.output_dict['total_erection_cost'].iterrows():
            dashed_row = '{} - {} - {}'.format(row.iloc[0], row.iloc[1], row.iloc[2])
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'total_erection_cost: Phase of construction - Type of cost - Cost USD',
                'value': dashed_row,
                'last_number': row.iloc[2]
            })

        for _, row in self.output_dict['erection_selected_detailed_data'].iterrows():
            value = row['Labor cost USD without management']
            operation = row['Operation']
            result.append({
                'unit': 'usd',
                'type': 'dataframe',
                'variable_df_key_col_name': 'erection_selected_detailed_data: crew cost without management',
                'value': value,
                'non_numeric_value': operation
            })

        for _, row in self.output_dict['erection_selected_detailed_data'].iterrows():
            value = row['Mobilization cost USD']
            crane_boom_operation_concat = row['crane_boom_operation_concat']
            result.append({
                'unit': 'usd',
                'type': 'dataframe',
                'variable_df_key_col_name': 'erection_selected_detailed_data: mobilization',
                'value': value,
                'non_numeric_value': crane_boom_operation_concat
            })

        for _, row in self.output_dict['erection_selected_detailed_data'].iterrows():
            value = row['Wind multiplier']
            operation = row['Operation']
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': f'erection_selected_detailed_data: wind multiplier',
                'value': value,
                'non_numeric_value': operation
            })

        result.append({
            'unit': 'usd',
            'type': 'variable',
            'variable_df_key_col_name': 'total_cost_summed_erection',
            'value': float(self.output_dict['total_cost_summed_erection'])
        })

        for _, row in self.output_dict['management_crews_cost'].iterrows():
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'management_crews_cost: {}'.format(' - '.join(row.index)),
                'value': ' - '.join(list(str(x) for x in row)[1:])
            })

        result.append({
            'unit': 'hours',
            'type': 'variable',
            'variable_df_key_col_name': 'number of hours in weather window',
            'value': len(self.input_dict['weather_window'])
        })

        result.append({
            'unit': 'none',
            'type': 'variable',
            'variable_df_key_col_name': 'time_weighted_weather_multiplier',
            'value': self.output_dict['time_weighted_weather_multiplier']
        })

        result.append({
            'unit': 'months',
            'type': 'variable',
            'variable_df_key_col_name': 'erection_construction_months',
            'value': self.output_dict['erection_construction_months']
        })

        result.append({
            'unit': 'usd',
            'type': 'variable',
            'variable_df_key_col_name': 'labor_cost_management',
            'value': self.output_dict['labor_cost_management']
        })

        result.append({
            'unit': 'usd',
            'type': 'variable',
            'variable_df_key_col_name': 'labor_cost_non_management',
            'value': self.output_dict['labor_cost_non_management']
        })

        result.append({
            'unit': 'usd',
            'type': 'variable',
            'variable_df_key_col_name': 'labor_cost_total',
            'value': self.output_dict['labor_cost_total']
        })


        module = type(self).__name__
        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module
        self.output_dict['erection_cost_csv'] = result

        return result

    def calculate_erection_operation_time(self):
        """
        Calculates operation time required for each type of equipment included in project data.

        self.output_dict['possible_cranes'] = possible_cranes
        self.output_dict['erection_operation_time'] = erection_operation_time_dict

        self.input_dict keys
        ---------------------
        construct_duration : int
            int duration of construction (in months)

        operational_construction_time : int
            Number of hours each day that are available for construction hours.

        self.output_dict keys
        ---------------------
        self.output_dict['possible_cranes'] : possible_cranes (with geometry)
        self.output_dict['erection_operation_time'] : Operation time for each crane.

        Returns
        -------
        pd.DataFrame, pd.DataFrame
            Dataframe of possible_cranes (with geometry) and operational time for cranes
        """
        project_data = self.input_dict['project_data']
        construct_duration = self.input_dict['construct_duration'] # Total project construction time (months)
        operational_construction_time = self.input_dict['operational_construction_time'] # Hours, 10 or 24

        # Why multiply by hardcoded 1/3?
        erection_construction_time = 1 / 3 * construct_duration
        breakpoint_between_base_and_topping_percent = self.input_dict['breakpoint_between_base_and_topping_percent']
        hub_height_m = self.input_dict['hub_height_meters']
        rotor_diameter_m = self.input_dict['rotor_diameter_m']
        num_turbines = float(self.input_dict['num_turbines'])
        turbine_spacing_rotor_diameters = self.input_dict['turbine_spacing_rotor_diameters']

        # for components in component list determine if base or topping
        project_data['components']['Operation'] = project_data['components']['Lift height m'] > (
            float(hub_height_m * breakpoint_between_base_and_topping_percent))
        boolean_dictionary = {True: 'Top', False: 'Base'}
        project_data['components']['Operation'] = project_data['components']['Operation'].map(boolean_dictionary)

        # For output to a csv file
        self.output_dict['component_name_topvbase'] = project_data['components'][['Component', 'Operation']]

        # create groups for operations
        top_v_base = project_data['components'].groupby('Operation')

        # group crane data by boom system and crane name to get distinct cranes
        crane_grouped = project_data['crane_specs'].groupby(
            ['Equipment name',  'Equipment ID', 'Crane name', 'Boom system', 'Crane capacity tonne'])

        # Calculate the crane lift polygons
        crane_poly = self.calculate_crane_lift_polygons(crane_grouped=crane_grouped)

        # loop through operation type (topping vs. base)
        component_max_speed = pd.DataFrame()
        for name_operation, component_group in top_v_base:
            lift_max_wind_speed = self.calculate_component_lift_max_wind_speed(component_group=component_group,
                                                                               crane_poly=crane_poly,
                                                                               component_max_speed=component_max_speed,
                                                                               operation=name_operation)
            crane_poly = lift_max_wind_speed['crane_poly']
            component_max_speed = lift_max_wind_speed['component_max_speed']

        # Sorting can help for some operations, but isn't strictly necessary, so it can be turned
        # off when not debugging
        # component_max_speed = component_max_speed.sort_values(by=['Crane name', 'Boom system', 'Component'])

        # join crane polygon to crane specs
        crane_component = pd.merge(crane_poly, component_max_speed, on=['Crane name', 'Boom system'], sort=True)

        # select only cranes that could lift the component
        possible_cranes = crane_component.where(crane_component['crane_bool'] == True).dropna(thresh=1).reset_index(
            drop=True)

        # calculate travel time per cycle
        turbine_spacing = float(
            turbine_spacing_rotor_diameters * rotor_diameter_m * km_per_m)
        possible_cranes['Travel time hr'] = turbine_spacing / possible_cranes['Speed of travel km per hr'] * num_turbines

        # calculate erection time
        possible_cranes['Operation time hr'] = ((possible_cranes['Lift height m'] / possible_cranes[
            'Hoist speed m per min'] * hr_per_min)
                                                + (possible_cranes['Cycle time installation hrs'])
                                                ) * num_turbines

        # Modify the breakdown time column to reflect all the crane breakdowns
        # the entire project
        crane_breakdown_fraction = self.input_dict['crane_breakdown_fraction']
        num_turbines_needing_breakdowns = ceil(num_turbines * crane_breakdown_fraction)
        breakdown_time_all_turbines_hrs = possible_cranes['Breakdown time hr'] * num_turbines_needing_breakdowns

        # Combine the breakdown time with the setup time to get total setup + breakdown
        # time and store it as setup time.
        setup_time_all_turbines_hrs = possible_cranes['Setup time hr'] * num_turbines
        possible_cranes['Setup time hr'] = setup_time_all_turbines_hrs + breakdown_time_all_turbines_hrs

        # check that crane can lift all components within a group (base vs top)
        crane_lift_entire_group_for_operation = crane_component.groupby(by=['Crane name', 'Boom system', 'Operation'])[
            'crane_bool'].all()
        # if it can't then we need to remove it.
        # otherwise we end up with an option for a crane to perform an operation without lifting all of the corresponding components
        testcranenew = possible_cranes.merge(crane_lift_entire_group_for_operation,
                                             on=['Crane name', 'Boom system', 'Operation'], sort=True)
        possible_cranes = testcranenew.loc[testcranenew['crane_bool_y']]

        erection_time = possible_cranes.groupby(['Crane name', 'Equipment name', 'Crane capacity tonne', 'Crew type ID',
                                                 'Boom system', 'Operation'])['Operation time hr'].sum()
        travel_time = possible_cranes.groupby(['Crane name', 'Equipment name', 'Crane capacity tonne', 'Crew type ID',
                                               'Boom system', 'Operation'])['Travel time hr'].max()
        setup_time = possible_cranes.groupby(['Crane name', 'Equipment name', 'Crane capacity tonne', 'Crew type ID',
                                              'Boom system', 'Operation'])['Setup time hr'].max()
        rental_time_without_weather = erection_time + travel_time + setup_time

        operation_time = rental_time_without_weather.reset_index()
        operation_time = operation_time.rename(columns={0: 'Operation time all turbines hrs'})
        operation_time['Operational construct days'] = (operation_time['Operation time all turbines hrs'] /
                                                        operational_construction_time)

        # if more than one crew needed to complete within construction duration then assume that all construction happens
        # within that window and use that time frame for weather delays; if not, use the number of days calculated
        operation_time['time_construct_bool'] = (operation_time['Operational construct days'] >
                                                 erection_construction_time * 30)
        boolean_dictionary = {True: erection_construction_time * 30, False: np.NAN}
        operation_time['time_construct_bool'] = operation_time['time_construct_bool'].map(boolean_dictionary)
        operation_time['Time construct days'] = operation_time[
            ['time_construct_bool', 'Operational construct days']].min(axis=1)

        # print(possible_cranes[['Crane name', 'Component', 'Operation time hr', 'Operation']])

        for operation, component_group in top_v_base:
            unique_component_crane = possible_cranes.loc[possible_cranes['Operation'] == operation][
                'Component'].unique()
            for component in component_group['Component']:
                if component not in unique_component_crane:
                    raise Exception(
                        'Error: Unable to find installation crane for {} operation and {} component'.format(operation,
                                                                                                            component))

        erection_operation_time_dict = dict()
        erection_operation_time_dict['possible_cranes'] = possible_cranes
        erection_operation_time_dict['operation_time'] = operation_time

        self.output_dict['possible_cranes'] = possible_cranes
        self.output_dict['erection_operation_time'] = erection_operation_time_dict

        return possible_cranes, operation_time

    def calculate_offload_operation_time(self):
        """
        Calculates time for the offload operation.

        self.input_dict keys
        --------------------
        project_data : dict
            dict of data frames for each of the csv files loaded for the project

        operational_construction_time : int
            operational hours of construction

        rate_of_deliveries : int
            rate of deliveries of turbines ready for erection.

        self.output_dict key
        --------------------
        possible_cranes : pd.DataFrame
            Dataframe of cranes possibly available for the operation

        operation_time : int
            Integer of number of hours per day construction can proceed.
        """
        project_data = self.input_dict['project_data']
        operational_construction_time = self.input_dict['operational_construction_time']
        rate_of_deliveries = self.input_dict['rate_of_deliveries']
        rotor_diameter_m = self.input_dict['rotor_diameter_m']
        num_turbines = float(self.input_dict['num_turbines'])
        turbine_spacing_rotor_diameters = self.input_dict['turbine_spacing_rotor_diameters']

        offload_cranes = project_data['crane_specs'].where(
            project_data['crane_specs']['Equipment name'] == 'Offload crane')

        # group crane data by boom system and crane name to get distinct cranes
        crane_grouped = offload_cranes.groupby(
            ['Equipment name', 'Equipment ID', 'Crane name', 'Boom system', 'Crane capacity tonne'])

        crane_poly = self.calculate_crane_lift_polygons(crane_grouped=crane_grouped)
        component_group = project_data['components']
        component_max_speed = pd.DataFrame()
        lift_max_wind_speed = self.calculate_component_lift_max_wind_speed(component_group=component_group,
                                                                           crane_poly=crane_poly,
                                                                           component_max_speed=component_max_speed,
                                                                           operation='offload')
        component_max_speed = lift_max_wind_speed['component_max_speed']
        crane_poly = lift_max_wind_speed['crane_poly']

        if len(crane_poly) != 0:
            # join crane polygon to crane specs
            crane_component = pd.merge(crane_poly, component_max_speed, on=['Crane name', 'Boom system'], sort=True)

            # select only cranes that could lift the component
            possible_cranes = crane_component.where(crane_component['crane_bool'] == True).dropna(thresh=1).reset_index(
                drop=True)

            # calculate travel time per cycle
            turbine_spacing = float(
                turbine_spacing_rotor_diameters * rotor_diameter_m * km_per_m)
            turbine_num = float(self.input_dict['num_turbines'])
            possible_cranes['Travel time hr'] = turbine_spacing / possible_cranes['Speed of travel km per hr'] * num_turbines

            # calculate erection time
            possible_cranes['Operation time hr'] = ((possible_cranes['Lift height m'] / possible_cranes[
                'Hoist speed m per min'] * hr_per_min)
                                                    + (possible_cranes['Offload cycle time hrs'])
                                                    ) * turbine_num

            # store setup time
            possible_cranes['Setup time hr'] = possible_cranes['Setup time hr'] * turbine_num

            erection_time = \
            possible_cranes.groupby(['Crane name', 'Equipment name', 'Crane capacity tonne', 'Crew type ID',
                                     'Boom system'])['Operation time hr'].sum()
            travel_time = \
            possible_cranes.groupby(['Crane name', 'Equipment name', 'Crane capacity tonne', 'Crew type ID',
                                     'Boom system'])['Travel time hr'].max()
            setup_time = \
            possible_cranes.groupby(['Crane name', 'Equipment name', 'Crane capacity tonne', 'Crew type ID',
                                     'Boom system'])['Setup time hr'].max()
            rental_time_without_weather = erection_time + travel_time + setup_time

            operation_time = rental_time_without_weather.reset_index()
            operation_time = operation_time.rename(columns={0: 'Operation time all turbines hrs'})
            operation_time['Operational construct days'] = (operation_time['Operation time all turbines hrs'] /
                                                            operational_construction_time)

            # if more than one crew needed to complete within construction duration
            # then assume that all construction happens within that window and use
            # that timeframe for weather delays; if not, use the number of days calculated
            operation_time['time_construct_bool'] = (turbine_num / operation_time['Operational construct days'] * 6
                                                     > float(rate_of_deliveries))
            boolean_dictionary = {True: (float(turbine_num) / (float(rate_of_deliveries) / 6)), False: np.NAN}
            operation_time['time_construct_bool'] = operation_time['time_construct_bool'].map(boolean_dictionary)
            operation_time['Time construct days'] = operation_time[
                ['time_construct_bool', 'Operational construct days']].max(
                axis=1)

            possible_cranes['Operation'] = 'Offload'
            operation_time['Operation'] = 'Offload'
        else:
            possible_cranes = []
            operation_time = []

        # print(possible_cranes[['Crane name', 'Component', 'Operation time hr']])
        unique_components = project_data['components']['Component'].unique()
        unique_component_crane = possible_cranes['Component'].unique()
        for component in unique_components:
            if component not in unique_component_crane:
                raise Exception('Error: Unable to find offload crane for {}'.format(component))

        return possible_cranes, operation_time

    def calculate_crane_lift_polygons(self, crane_grouped):
        """
        Here we associate polygons with each crane. However, these polygons are not shapes
        for the lift. Rather, they define functions f(x), where x is a crane lift load and
        f(x) is the height to which that load can be lifted. To find out whether the crane
        can lift a particular load, one just needs to check whether a point x (lift mass in
        tonnes) and y (lift height in m) lies within the crane's polygon.

        Parameters
        ----------
        crane_grouped : pandas.core.groupby.generic.DataFrameGroupBy
            The aggregation of the cranes to compute the lift polygons for. The columns
            in the aggregation are assume to be 'Equipment name', 'Crane name', 'Boom system',
            'Crane capacity tonne'

        Returns
        -------
        pd.DataFrame
            A dataframe of the cranes and their lifting polygons.
        """
        crane_poly = pd.DataFrame(
            columns=['Equipment name', 'Equipment ID', 'Crane name', 'Boom system', 'Crane capacity tonne', 'Crane poly'])
        for (equipment_name, equipment_id, crane_name, boom_system, crane_capacity_tonne), crane in crane_grouped:
            crane = crane.reset_index(drop=True)
            x = crane['Max capacity tonne']
            y = crane['Hub height m']
            wind_speed = min(crane['Max wind speed m per s'])
            hoist_speed = min(crane['Hoist speed m per min'])
            travel_speed = min(crane['Speed of travel km per hr'])
            setup_time = max(crane['Setup time hr'])
            breakdown_time = max(crane['Breakdown time hr'])
            crew_type = crane.loc[0, 'Crew type ID'] # For every crane/boom combo the crew is the same, so we can just take first crew.
            polygon = [Point(0, 0), Point(0, max(y)), Point(min(x), max(y)), Point(max(x), min(y)), Point(max(x), 0)]
            df = pd.DataFrame([[equipment_name,
                                equipment_id,
                                crane_name,
                                boom_system,
                                crane_capacity_tonne,
                                wind_speed,
                                setup_time,
                                breakdown_time,
                                hoist_speed,
                                travel_speed,
                                crew_type,
                                polygon]],
                              columns=['Equipment name', 'Equipment ID', 'Crane name', 'Boom system', 'Crane capacity tonne',
                                       'Max wind speed m per s', 'Setup time hr', 'Breakdown time hr',
                                       'Hoist speed m per min', 'Speed of travel km per hr',
                                       'Crew type ID', 'Crane poly'])
            crane_poly = pd.concat((crane_poly, df), sort=True)
        return crane_poly

    def calculate_component_lift_max_wind_speed(self, *, component_group, crane_poly, component_max_speed, operation):
        """
        First, using the height and mass of the component being lifted, this method determines
        if a component can be lifted to the necessary height by each crane.

        Also, creates a dataframe that has the maximum wind speeds to lift particular components,
        given the component data and crane lift data given in the arguments.

        For the maximum wind speed calculations, we use these equations to calculation vmax,
        which is the maximum permissible wind speed:

        vmax = max_TAB * sqrt(1.2 * mh / aw), where
        mh = hoist load
        aw = area exposed to wind = surface area * coeff drag
        1.2 = constant in m^2 / t
        vmax_tab = maximum load speed per load chart
        (source: pg. 33 of Liebherr)

        See the source code for this method on how this calculation is used.

        Parameters
        ----------
        component_group : pd.DataFrame
            Dataframe with component data.

        crane_poly : pd.DataFrame
            Data about cranes doing the lifting. The polygons are specifications of
            functions that define lift height f(x) as a function of component mass x.

        component_max_speed : pd.DataFrame
            The dataframe into which maximum wind speeds for lifting each component
            will be accumulated. For the first call into this method, pass in an
            empty dataframe created with pd.DataFrame

        operation : str
            The name of the operation ("base", "top" or "offload") that the cranes
            are performing for this calculation. If the operation is "Offload"
            the 'Mass tonne' is divided by two when making the lift polygons.
            This created the assumption that there are always 2 offload cranes during
            offload operations. (See the calculate_crane_lift_polygons() method
            above for more about calculating the lift polygons.

        Returns
        -------
        dict
            Returns a dict of pd.DataFrame values. The key "component_max_speed" is the
            the dataframe of max component speeds. The key "crane_poly" is a COPY of the
            crane_poly dataframe passed as a parameter to this function and with a column
            of "Crane bool {operation}" attached.
        """
        for idx, crane in crane_poly.iterrows():
            polygon = crane['Crane poly']

            # calculate polygon for crane capacity and check if component can be lifted by each crane without wind loading
            for component in component_group['Component']:
                # get weight and height of component in each component group
                component_only = component_group.where(component_group['Component'] == component).dropna(thresh=1)

                # See docstring for "operation" parameter above about mass calculations for offloading
                if operation == 'offload':
                    point = Point(component_only['Mass tonne'] / 2, (component_only['Section height m'] + component_only['Offload hook height m']))
                else:
                    point = Point(component_only['Mass tonne'], (component_only['Lift height m'] + component_only['Offload hook height m']))
                crane['Lift boolean {component}'.format(component=component)] = point_in_polygon(point, polygon)

            # Transform the "Lift boolean" indexes in the series to a list of booleans
            # that signify if the crane can lift a component.
            bool_list = []
            for component in component_group['Component']:
                if crane['Lift boolean {component}'.format(component=component)] is False:
                    crane_bool = False
                else:
                    crane_bool = True
                bool_list.append(crane_bool)

            # mh is an effective mass (it should be the mass of the entire component for both offload and other cranes, not just 1/2 that's used above for determining whether the part can be lifted)
            mh = component_group['Mass tonne']
            aw = component_group['Surface area sq m'] * component_group['Coeff drag']
            vmax_tab = crane['Max wind speed m per s']
            vmax_calc = vmax_tab * np.sqrt(1.2 * mh / aw)

            # if vmax_calc is less than vmax_tab then vmax_calc, otherwise vmax_tab (based on pg. 33 of Liebherr)
            component_group_new = pd.DataFrame(component_group,
                                               columns=list(component_group.columns.values) + ['vmax',
                                                                                               'Crane name',
                                                                                               'Boom system',
                                                                                               'crane_bool'])
            component_group_new['vmax'] = np.minimum(vmax_tab, vmax_calc)
            component_group_new['Crane name'] = crane['Crane name']
            component_group_new['Boom system'] = crane['Boom system']
            component_group_new['crane_bool'] = bool_list

            component_max_speed = pd.concat((component_max_speed, component_group_new), sort=True)

        crane_poly_new = crane_poly.copy()
        crane_poly_new['Crane bool {}'.format(operation)] = min(bool_list)

        result = {
            'component_max_speed': component_max_speed,
            'crane_poly': crane_poly_new
        }

        return result

    def calculate_wind_delay_by_component(self):
        """
        Calculates wind delay for each component in the project.

        Returns
        -------
        pd.DataFrame
            crane specifications and component properties joined with wind delays for each case.
        """

        # Get necessary values from input_dict
        crane_specs = self.output_dict['crane_specs_withoffload']
        weather_window = self.input_dict['weather_window']

        # calculate wind delay for each component and crane combination
        crane_specs = crane_specs.reset_index()
        crane_specs['Wind delay percent'] = np.nan

        # pull global inputs for weather delay from input_dict
        weather_data_keys = {'wind_shear_exponent', 'weather_window'}

        # specify collection-specific weather delay inputs
        weather_delay_global_inputs = {i: self.input_dict[i] for i in self.input_dict if i in weather_data_keys}

        # Iterate over every crane + boom combination
        for i, row in crane_specs.iterrows():
            # assume we don't know when the operation occurs
            operation_window = len(weather_window.index)  # operation window = entire construction weather window
            operation_start = 0  # start time is at beginning of construction weather window

            # extract critical wind speed
            critical_wind_operation = row['vmax']

            # extract height of interest (differs for offload cranes)
            if (row['Crane bool offload'] == 1) is True:
                height_interest = row['Section height m'] + row['Offload hook height m']
            else:
                height_interest = row['Lift height m'] + row['Offload hook height m']

            # compute weather delay
            weather_delay_input_dict = weather_delay_global_inputs
            weather_delay_output_dict = dict()
            weather_delay_input_dict['start_delay_hours'] = operation_start
            weather_delay_input_dict['critical_wind_speed_m_per_s'] = critical_wind_operation
            weather_delay_input_dict['wind_height_of_interest_m'] = height_interest
            weather_delay_input_dict['mission_time_hours'] = operation_window

            WeatherDelay(weather_delay_input_dict, weather_delay_output_dict)
            wind_delay = np.array(weather_delay_output_dict['wind_delays'])

            # if greater than 4 hour delay, then shut down for full day (10 hours)
            wind_delay[(wind_delay > 4)] = 10
            wind_delay_time = float(wind_delay.sum())

            # store weather delay for operation, component, crane, and boom combination
            crane_specs.loc[i, 'Wind delay percent'] = wind_delay_time / len(weather_window)

        self.output_dict['enhanced_crane_specs'] = crane_specs
        return crane_specs

    def aggregate_erection_costs(self):
        """
        Aggregates labor, equipment, mobilization and fuel costs for erection.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]
            Two dataframes: First, utilizing the separate cranes for base and topping.
            Second, utilizing same crane for base and topping,
            Third is crew cost.
        """
        join_wind_operation = self.output_dict['join_wind_operation']
        overtime_multiplier = self.input_dict['overtime_multiplier']
        time_construct = self.input_dict['time_construct']
        project_data = self.input_dict['project_data']
        hour_day = self.input_dict['hour_day']

        # TODO: consider removing equipment name and crane capacity from crane_specs tab (I believe these data are unused here and they get overwritten later with equip information from equip tab)
        join_wind_operation = join_wind_operation.drop(columns=['Equipment name', 'Crane capacity tonne'])

        possible_crane_cost_with_equip = pd.merge(join_wind_operation, project_data['equip'],
                                                  on=['Equipment ID', 'Operation'], sort=True)

        equip_crane_cost = pd.merge(possible_crane_cost_with_equip, project_data['equip_price'],
                                       on=['Equipment name', 'Crane capacity tonne'], sort=True)

        equip_crane_cost['Equipment rental cost USD'] = equip_crane_cost['Total time per op with weather'] * \
                                                        equip_crane_cost['Equipment price USD per hour'] * \
                                                        equip_crane_cost['Number of equipment']

        # Drop duplicates from equip crane cost, if any
        # equip_crane_cost.drop_duplicates(subset=['Equipment ID', 'Operation', 'Crane name', 'Boom system'], inplace=True)

        equipment_cost_to_merge = equip_crane_cost[['Crane name', 'Boom system', 'Equipment ID', 'Operation', 'Equipment price USD per hour', 'Number of equipment', 'Equipment rental cost USD', 'Fuel consumption gal per day']]
        equipment_cost_to_merge = equipment_cost_to_merge.groupby(['Crane name', 'Boom system', 'Equipment ID', 'Operation']).sum().reset_index()

        possible_crane_cost = pd.merge(join_wind_operation, equipment_cost_to_merge, on=['Crane name', 'Boom system', 'Equipment ID', 'Operation'], sort=True)

        # Remove any duplicates from crew data.
        crew_deduped = project_data['crew'].drop_duplicates(
            subset=['Crew type ID', 'Operation', 'Crew name', 'Labor type ID'], keep="first")

        # Merge crew and price data for non-management crews only (base, topping, and offload only)
        crew_cost = pd.merge(crew_deduped, project_data['crew_price'], on=['Labor type ID'], sort=True)
        self.output_dict['crew_cost'] = crew_cost
        non_management_crew_cost = crew_cost.loc[crew_cost['Operation'].isin(['Base', 'Top', 'Offload'])]

        # calculate crew costs
        non_overtime_hours_per_week = 40
        working_days_per_week = 6
        hours_per_week = working_days_per_week * hour_day[time_construct]
        overtime_percentage = (hours_per_week - non_overtime_hours_per_week) / hours_per_week
        normal_labor_rate = non_overtime_hours_per_week / hours_per_week
        crew_cost['Hourly rate for all workers'] = (non_management_crew_cost['Hourly rate USD per hour'] * non_management_crew_cost[
            'Number of workers']) * (normal_labor_rate + overtime_percentage * overtime_multiplier)
        crew_cost['Per diem all workers'] = non_management_crew_cost['Per diem USD per day'] * non_management_crew_cost['Number of workers']

        # Crew cost group is getting two more people in it.
        # Note to future self, enforce constraints on dataframes.
        #
        # Before crew_price sheet is used, sort based on labor cost. Then drop rows with duplicated job titles.
        # Intent is to keep the most expensive labor row.

        # group crew costs by crew type and operation
        crew_cost_grouped = crew_cost.groupby(['Crew type ID', 'Operation']).sum(numeric_only=True).reset_index()

        # merge crane data with grouped crew costs

        possible_crane_cost = pd.merge(possible_crane_cost, crew_cost_grouped, on=['Crew type ID', 'Operation'], sort=True)

        # calculate labor costs
        labor_day_operation = round(possible_crane_cost['Total time per op with weather'] / hour_day[time_construct])
        possible_crane_cost['Subtotal for hourly labor (non-management) USD'] = possible_crane_cost['Total time per op with weather'] * possible_crane_cost['Hourly rate for all workers']
        possible_crane_cost['Subtotal for per diem labor (non-management) USD'] = labor_day_operation * possible_crane_cost['Per diem all workers']
        possible_crane_cost['Labor cost USD without management'] = possible_crane_cost['Subtotal for hourly labor (non-management) USD'] + \
                                                                   possible_crane_cost['Subtotal for per diem labor (non-management) USD']


        # calculate fuel costs
        possible_crane_cost['Fuel cost USD'] = possible_crane_cost['Fuel consumption gal per day'] * float(
            self.input_dict['fuel_cost_usd_per_gal']) * labor_day_operation

        # calculate costs if top and base cranes are the same
        base_cranes = possible_crane_cost[possible_crane_cost['Operation'] == 'Base']
        top_cranes = possible_crane_cost[possible_crane_cost['Operation'] == 'Top']
        crane_topbase_bool = top_cranes['Crane name'].isin(base_cranes['Crane name'])
        boom_topbase_bool = top_cranes['Boom system'].isin(base_cranes['Boom system'])
        possible_crane_topbase = top_cranes[boom_topbase_bool & crane_topbase_bool]
        same_topbase_crane_list = possible_crane_topbase[['Crane name', 'Boom system']]
        possible_crane_topbase  = same_topbase_crane_list.merge(possible_crane_cost, on=['Crane name', 'Boom system'], sort=True)
        possible_crane_topbase_sum = possible_crane_topbase.groupby(['Crane name',
                                                                     'Boom system'])[['Labor cost USD without management',
                                                                                     'Subtotal for hourly labor (non-management) USD',
                                                                                     'Subtotal for per diem labor (non-management) USD',
                                                                                     'Equipment rental cost USD',
                                                                                     'Fuel cost USD']
        ].sum().reset_index()

        # Store the possible cranes for the top and base for future diagnostics.
        self._possible_crane_cost = possible_crane_cost.copy()

        # group crane spec data for mobilization
        mobilization_costs = project_data['crane_specs'].groupby(['Crane name', 'Boom system'])[
            'Mobilization cost USD'].max().reset_index()

        mobilization_costs['Mobilization cost USD'] = mobilization_costs['Mobilization cost USD'] * 2 # for mobilization and demobilizaton

        # join top and base crane data with mobilization data
        topbase_same_crane_cost = pd.merge(possible_crane_topbase_sum, mobilization_costs,
                                           on=['Crane name', 'Boom system'], sort=True)

        # compute total project cost for erection
        topbase_same_crane_cost['Total cost USD'] = topbase_same_crane_cost['Labor cost USD without management'] + \
                                                    topbase_same_crane_cost['Equipment rental cost USD'] + \
                                                    topbase_same_crane_cost['Fuel cost USD'] + \
                                                    topbase_same_crane_cost[
                                                        'Mobilization cost USD']

        # adds operation label for same crane used for base and topping (this way columns are consistent for same and separate basetop)
        topbase_same_crane_cost['Operation'] = 'Base + Top'

        # calculate costs if top and base use separate cranes
        separate_topbase = \
            possible_crane_cost.groupby(['Operation', 'Crane name', 'Boom system'])[['Labor cost USD without management',
                                                                                                   'Subtotal for hourly labor (non-management) USD',
                                                                                                   'Subtotal for per diem labor (non-management) USD',
                                                                                                   'Equipment rental cost USD',
                                                                                                   'Fuel cost USD']
        ].sum().reset_index()

        # join mobilization data to separate top base crane costs
        separate_topbase_crane_cost = pd.merge(separate_topbase, mobilization_costs, on=['Crane name', 'Boom system'], sort=True)

        # compute total project cost for erection
        separate_topbase_crane_cost['Total cost USD'] = separate_topbase_crane_cost['Labor cost USD without management'] + \
                                                        separate_topbase_crane_cost['Equipment rental cost USD'] + \
                                                        separate_topbase_crane_cost['Fuel cost USD'] + \
                                                        separate_topbase_crane_cost[
                                                            'Mobilization cost USD']

        return separate_topbase_crane_cost, topbase_same_crane_cost, crew_cost


    def find_minimum_cost_cranes(self):
        """
        Finds the minimum cost crane(s) based on the aggregated labor, equipment,
        mobilization and fuel costs for erection.

        self.output_dict keys used as inputs
        ------------------------------------
        separate_basetop : pd.DataFrame
            data frame with aggregated labor, equipment, mobilization, and fuel costs for utilizing
            separate cranes for base and topping.
        same_basetop : pd.DataFrame
            data frame with aggregated labor, equipment, mobilization, and fuel costs for utilizing the
            same crane for base and topping

        allow_same_flag : boolean
            flag to indicate whether choosing same base and topping crane is allowed
        """
        allow_same_flag = self.input_dict['allow_same_flag']
        separate_basetop = self.output_dict['separate_basetop']
        same_basetop = self.output_dict['same_basetop']

        self.output_dict['separate_basetop'] = separate_basetop

        total_separate_cost = pd.DataFrame()
        for operation in separate_basetop['Operation'].unique():
            # find minimum cost option for separate base and topping cranes
            min_val = min(separate_basetop['Total cost USD'].where(separate_basetop['Operation'] == operation).dropna())

            # find the crane that corresponds to the minimum cost for each operation
            crane = separate_basetop[separate_basetop['Total cost USD'] == min_val]
            cost = crane.groupby('Operation').min()
            total_separate_cost = pd.concat((total_separate_cost, cost), sort=True)

        # reset index for separate crane costs
        total_separate_cost = total_separate_cost.reset_index()

        # duplicate offload records because assuming two offload cranes are on site
        total_separate_cost = pd.concat((total_separate_cost,
                                         total_separate_cost.loc[total_separate_cost['Operation'] == 'Offload']),
                                        sort=True)

        # sum costs for separate cranes to get total for all cranes
        cost_chosen_separate = total_separate_cost['Total cost USD'].sum()

        if allow_same_flag is True:
            # get the minimum cost for using the same crane for all operations
            cost_chosen_same = min(same_basetop['Total cost USD'])

            # check if separate or same crane option is cheaper and choose crane cost
            if cost_chosen_separate < cost_chosen_same:
                cost_chosen = total_separate_cost.groupby(by=["Boom system", "Crane name", "Operation"]).sum()  # added crane name and operation to groupby
            else:
                cost_chosen = same_basetop.where(same_basetop['Total cost USD'] == cost_chosen_same).dropna()
        else:
            cost_chosen = total_separate_cost.groupby(by=["Boom system", "Crane name", "Operation"]).sum()  # added crane name and operation to groupby

        return cost_chosen

    def calculate_management_crews_cost(self, erection_cost):
        """
        Calculates management costs for erection, based on rate of turbine deliveries.

        Parameters
        ----------
        erection_cost : pd.DataFrame
            Other erection costs from which to calculate duration of management
            operations.

        Returns
        -------
        pd.DataFrame, pd.DataFrame, float
            The first dataframe are management costs by each role on each team.
            The second dataframe are management costs summed over aggregations of
            teams. The float is the sum of all the management costs for erection.
        """

        num_turbines = self.input_dict['num_turbines']
        overtime_multiplier = self.input_dict['overtime_multiplier']
        hour_day = self.input_dict['hour_day']
        time_construct = self.input_dict['time_construct']
        deliveries_per_week = float(self.input_dict['rate_of_deliveries'])
        duration_days = erection_cost['Time construct days'].sum()  
        duration_hours = round(duration_days * hour_day[time_construct])

        # Merge crew and price data for management crews only.
        crew_cost = self.output_dict['crew_cost']
        management_crews = crew_cost.loc[crew_cost['Operation'].isin(['Management', 'Mechanical completion'])].copy()

        # increase management crews by project size
        management_crews.loc[management_crews['Crew name'] == 'Management - project size', 'Number of workers'] = \
            round(management_crews.loc[management_crews['Crew name'] == 'Management - project size', 'Number of workers'] *
                  np.ceil(num_turbines / 100))

        # increase management crews by rate of construction (scale if greater than 10/wk)
        management_crews.loc[management_crews['Crew name'] == 'Management - rate construction', 'Number of workers'] = \
            round(management_crews.loc[management_crews['Crew name'] == 'Management - rate construction', 'Number of workers'] *
                  np.ceil(deliveries_per_week / 10))
        management_crews.loc[management_crews['Crew name'] == 'Mechanical completion', 'Number of workers'] = \
            round(management_crews.loc[management_crews['Crew name'] == 'Mechanical completion', 'Number of workers'] *
                  np.ceil(deliveries_per_week / 10))

        management_crews['Hourly rate for all workers'] = (management_crews['Hourly rate USD per hour'] * \
                                                           management_crews['Number of workers'])

        management_crews['Per diem all workers'] = management_crews['Per diem USD per day'] * \
                                                   management_crews['Number of workers']

        # Now calculate management costs
        management_crews['per_diem_costs'] = management_crews['Per diem all workers'] * duration_days
        hours_per_week = 6 * hour_day[time_construct]
        overtime_percentage = (hours_per_week - 40) / hours_per_week
        normal_labor_rate = 40 / hours_per_week
        management_crews['hourly_costs'] = management_crews['Hourly rate for all workers'] * (duration_hours * (normal_labor_rate + overtime_percentage * overtime_multiplier))
        management_crews['crew_level_total_costs'] = management_crews['per_diem_costs'] + \
                                                     management_crews['hourly_costs']

        # Aggregate and sum
        management_crew_cost_grouped = \
            management_crews.groupby(['Crew type ID', 'Operation', 'Crew name']).sum(numeric_only=True).reset_index()

        # Total management cost
        total_management_cost = management_crews['crew_level_total_costs'].sum()

        # erection_construction_months is the duration of erection time in
        # units of months.
        days_per_month = 30
        self.output_dict['erection_construction_months'] = duration_days / days_per_month

        return management_crews, management_crew_cost_grouped, total_management_cost

    def calculate_costs(self):
        """
        Calculates BOS costs for erection including selecting cranes that can lift
        components, incorporating wind delays and finding the least cost crane options
        for erection.
        """
        [crane_specs, operation_time] = self.calculate_erection_operation_time()
        crane_specs = crane_specs.infer_objects() # cast cols of Booleans to bool dtype

        self.output_dict['crane_specs'] = crane_specs
        self.output_dict['operation_time'] = operation_time

        [offload_specs, offload_time] = self.calculate_offload_operation_time()
        offload_specs = offload_specs.infer_objects() # cast cols of Booleans to bool dtype

        self.output_dict['offload_specs'] = offload_specs
        self.output_dict['offload_time'] = offload_time

        # append data for offloading
        if len(offload_specs) != 0:
            crane_specs_withoffload = pd.concat((crane_specs, offload_specs), sort=True)
            operation_time_withoffload = pd.concat((operation_time, offload_time), sort=True)
        else:
            raise Exception('ErectionCost calculate_costs(): offload_specs empty')

        self.output_dict['crane_specs_withoffload'] = crane_specs_withoffload
        self.output_dict['operation_time_withoffload'] = operation_time_withoffload
        crane_specs_with_weather = self.calculate_wind_delay_by_component()
        self.output_dict['cranes_wind_delay_withoffload'] = crane_specs_with_weather

        average_wind_delay = crane_specs_with_weather.groupby(['Crane name',
                                                               'Boom system',
                                                               'Operation',
                                                               'Equipment ID'])['Wind delay percent'].mean().reset_index()

        join_wind_operation = pd.merge(operation_time_withoffload, average_wind_delay,
                                       on=['Crane name', 'Boom system', 'Operation'], sort=True)

        join_wind_operation['Wind multiplier'] = 1 / (1 - join_wind_operation['Wind delay percent'])

        # 'Total time per op with weather' units are hours
        join_wind_operation['Total time per op with weather'] = (join_wind_operation['Operation time all turbines hrs'] *
                                                                 join_wind_operation['Wind multiplier'])

        self.output_dict['join_wind_operation'] = join_wind_operation

        [separate_basetop, same_basetop, crew_cost] = self.aggregate_erection_costs()

        self.output_dict['separate_basetop'] = separate_basetop
        self.output_dict['same_basetop'] = same_basetop
        self.output_dict['crew_cost'] = crew_cost

        erection_cost = self.find_minimum_cost_cranes()

        selected_time = join_wind_operation[['Crane name', 'Boom system', 'Operation', 'Operation time all turbines hrs', 'Total time per op with weather', 'Wind multiplier',
       'Operational construct days', 'Time construct days']]
        erection_cost = erection_cost.reset_index()
        selected_detailed_data = erection_cost.merge(selected_time, on=['Crane name', 'Boom system', 'Operation'], sort=True)
        selected_detailed_data['Total time per turbine'] = selected_detailed_data['Total time per op with weather'] / self.input_dict['num_turbines']

        management_crews_cost, management_crews_cost_grouped, total_management_cost = self.calculate_management_crews_cost(selected_detailed_data)

        crane_choice = selected_detailed_data[['Crane name', 'Boom system', 'Operation']].drop_duplicates()

        selected_detailed_data['crane_boom_operation_concat'] = selected_detailed_data[['Crane name', 'Boom system', 'Operation']].apply(lambda x: '-'.join(x), axis=1)
        crane_data_output = selected_detailed_data.drop(['Crane name', 'Boom system', 'Operation'], axis=1)
        crane_data_output = crane_data_output.melt(id_vars=['crane_boom_operation_concat'])

        crane_cost_details = crane_data_output.where(crane_data_output['variable'].str.contains("cost")).dropna()
        crane_cost_details = crane_cost_details.rename(index=str,
                            columns={"crane_boom_operation_concat": "Operation ID", "variable": "Type of cost",
                                     "value": "Cost"})

        subtotal_per_diem_labor_management_USD = management_crews_cost['per_diem_costs'].sum()
        subtotal_hourly_labor_management_USD = management_crews_cost['hourly_costs'].sum()

        self.output_dict['labor_cost_management'] = subtotal_per_diem_labor_management_USD + subtotal_hourly_labor_management_USD
        self.output_dict['labor_cost_non_management'] = selected_detailed_data['Labor cost USD without management'].sum()
        self.output_dict['labor_cost_total'] = self.output_dict['labor_cost_management'] + self.output_dict['labor_cost_non_management']

        total_erection_cost = pd.DataFrame(
            [['Erection', 'Equipment rental', selected_detailed_data['Equipment rental cost USD'].sum()],
             ['Erection', 'Fuel', selected_detailed_data['Fuel cost USD'].sum()],
             ['Erection', 'Labor', self.output_dict['labor_cost_total']],
             ['Erection', 'Mobilization', selected_detailed_data['Mobilization cost USD'].sum()],
             ['Erection', 'Other', 0],
             ['Erection', 'Materials', 0]],
            columns=['Phase of construction', 'Type of cost', 'Cost USD'])

        total_cost_summed_erection = total_erection_cost.sum(numeric_only=True).iloc[0]

        erection_wind_mult = selected_detailed_data['Wind multiplier']
        erection_wind_mult = erection_wind_mult.reset_index(drop=True).mean()

        self.output_dict['total_erection_cost'] = total_erection_cost
        self.output_dict['erection_wind_mult'] = erection_wind_mult
        self.output_dict['crane_choice'] = crane_choice
        self.output_dict['crane_data_output'] = crane_data_output
        self.output_dict['crane_cost_details'] = crane_cost_details
        self.output_dict['total_cost_summed_erection'] = total_cost_summed_erection

        # Put some diagnostic data on selected_detailed_data. This is the number of crews needed
        # To complete the construction withing the construction duration.

        # Operational days over time construct days isn't necessarily a proxy for a particular value.
        # But in case it is useful for something, here it is.

        selected_detailed_data['Operational construct days over time construct days'] = \
            np.ceil(selected_detailed_data['Operational construct days'] / selected_detailed_data['Time construct days'])

        total_time_construct_days = (selected_detailed_data['Time construct days']).sum()
        self.output_dict['time_weighted_weather_multiplier'] = (selected_detailed_data['Wind multiplier'] * (selected_detailed_data['Time construct days']) / total_time_construct_days).sum()

        # Now get the number of equipment diagnostic data ready. This is held on an instance
        # attribute because it isn't meant to be used outside of the class.
        self._number_of_equip = selected_detailed_data.merge(self._possible_crane_cost, on=['Crane name', 'Boom system', 'Operation'], how='inner', sort=True)
        self._number_of_equip = self._number_of_equip[['Operation', 'Crane name', 'Boom system', 'Number of equipment']]

        # Management crews data
        self.output_dict['management_crews_cost'] = management_crews_cost
        self.output_dict['management_crews_cost_grouped'] = management_crews_cost_grouped
        self.output_dict['erection_selected_detailed_data'] = selected_detailed_data



import os
import unittest

from wisdem.glue_code.runWISDEM import run_wisdem

test_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
    + os.sep
    + "examples"
    + os.sep
    + "02_reference_turbines"
    + os.sep
)
fname_modeling_options = test_dir + "modeling_options.yaml"
fname_analysis_options = test_dir + "analysis_options.yaml"


class TestRegression(unittest.TestCase):
    def test5MW(self):
        ## NREL 5MW
        fname_wt_input = test_dir + "nrel5mw.yaml"

        wt_opt, _, _ = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)

        self.assertAlmostEqual(wt_opt["rotorse.rp.AEP"][0] * 1.0e-6, 23.9127972556, 2)
        self.assertAlmostEqual(
            wt_opt["rotorse.blade_mass"][0], 16419.8666989212, 2
        )  # new value: improved interpolation
        self.assertAlmostEqual(wt_opt["financese.lcoe"][0] * 1.0e3, 51.0854048435, 1)
        self.assertAlmostEqual(wt_opt["rotorse.rs.tip_pos.tip_deflection"][0], 4.7295143980, 1)
        self.assertAlmostEqual(wt_opt["towerse.z_param"][-1], 87.7, 2)

    def test15MW(self):
        ## IEA 15MW
        fname_wt_input = test_dir + "IEA-15-240-RWT.yaml"
        wt_opt, _, _ = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)

        self.assertAlmostEqual(wt_opt["rotorse.rp.AEP"][0] * 1.0e-6, 77.8576828154, 1)
        self.assertAlmostEqual(
            wt_opt["rotorse.blade_mass"][0], 68206.4068005262, 1
        )  # new value: improved interpolation
        self.assertAlmostEqual(wt_opt["financese.lcoe"][0] * 1.0e3, 84.4784585961, 1)
        self.assertAlmostEqual(wt_opt["rotorse.rs.tip_pos.tip_deflection"][0], 25.831825879332392, 1)
        self.assertAlmostEqual(wt_opt["towerse.z_param"][-1], 144.386, 3)

    def test3p4MW(self):
        ## IEA 3.4MW
        fname_wt_input = test_dir + "IEA-3p4-130-RWT.yaml"
        wt_opt, _, _ = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)

        self.assertAlmostEqual(wt_opt["rotorse.rp.AEP"][0] * 1.0e-6, 13.6216930312, 1)
        self.assertAlmostEqual(
            wt_opt["rotorse.blade_mass"][0], 14528.4450828773, 1
        )  # new value: improved interpolation
        self.assertAlmostEqual(wt_opt["financese.lcoe"][0] * 1.0e3, 37.0720142587, 1)
        self.assertAlmostEqual(wt_opt["rotorse.rs.tip_pos.tip_deflection"][0], 7.3530140897, 1)
        self.assertAlmostEqual(wt_opt["towerse.z_param"][-1], 108.0, 3)


if __name__ == "__main__":
    unittest.main()



"""Tests for the `Vessel` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

import pandas as pd
import pytest
from marmot import le

from wisdem.orbit.core import Environment
from wisdem.test.test_orbit.data import test_weather as _weather

# Weather data with 'windspeed' column
simple_weather = _weather.copy()

# Weather data with windspeed heights and complex varied column name precision
weather = pd.DataFrame(_weather)
weather["windspeed_10m"] = weather["windspeed"].copy()
weather = weather.drop("windspeed", axis=1)
weather["windspeed_100.0m"] = weather["windspeed_10m"].copy() * 1.3
weather["windspeed_1.2m"] = weather["windspeed_10m"].copy() * 0.7
weather = weather.to_records()


def test_simple_inputs():
    env = Environment(state=simple_weather)

    assert "waveheight" in env.state.dtype.names
    assert "windspeed" in env.state.dtype.names


def test_inputs():
    env = Environment(state=weather)

    assert "waveheight" in env.state.dtype.names
    assert "windspeed" not in env.state.dtype.names
    assert "windspeed_10m" in env.state.dtype.names
    assert "windspeed_100m" in env.state.dtype.names
    assert "windspeed_1.2m" in env.state.dtype.names

    env2 = Environment(state=weather, ws_alpha=0.12, ws_default_height=20)

    assert env2.alpha == 0.12
    assert env2.default_height == 20


def test_simple_constraint_application():
    env = Environment(state=simple_weather)

    constraints = {"windspeed": le(10)}
    valid = env._find_valid_constraints(**constraints)
    assert valid == constraints

    with_height = {"windspeed_100m": le(10)}
    valid = env._find_valid_constraints(**with_height)
    assert "windspeed" in list(valid.keys())

    with_mult_heights = {"windspeed_100m": le(10), "windspeed_10m": le(8)}
    with pytest.raises(ValueError):
        valid = env._find_valid_constraints(**with_mult_heights)


def test_constraint_application():
    env = Environment(state=weather)

    constraints = {"waveheight": le(2), "windspeed": le(10)}
    valid = env._find_valid_constraints(**constraints)
    assert "windspeed_10m" in list(valid.keys())

    constraints = {"waveheight": le(2), "windspeed_10m": le(10)}
    valid = env._find_valid_constraints(**constraints)
    assert "waveheight" in list(valid.keys())
    assert "windspeed_10m" in list(valid.keys())

    assert "windspeed_20m" not in env.state.dtype.names
    constraints = {"waveheight": le(2), "windspeed_20m": le(10)}
    valid = env._find_valid_constraints(**constraints)
    assert "windspeed_20m" in list(valid.keys())
    assert "windspeed_20m" in env.state.dtype.names

    assert "windspeed_120m" not in env.state.dtype.names
    constraints = {"waveheight": le(2), "windspeed_120m": le(10)}
    valid = env._find_valid_constraints(**constraints)
    assert "windspeed_120m" in list(valid.keys())
    assert "windspeed_120m" in env.state.dtype.names


def test_interp():
    env = Environment(state=weather)

    assert "windspeed_20m" not in env.state.dtype.names
    constraints = {"waveheight": le(2), "windspeed_20m": le(10)}
    valid = env._find_valid_constraints(**constraints)
    assert "windspeed_20m" in env.state.dtype.names
    assert (env.state["windspeed_10m"] < env.state["windspeed_20m"]).all()
    assert (env.state["windspeed_20m"] < env.state["windspeed_100m"]).all()


def test_extrap():
    env = Environment(state=weather)

    assert "windspeed_120m" not in env.state.dtype.names
    constraints = {"waveheight": le(2), "windspeed_120m": le(10)}
    valid = env._find_valid_constraints(**constraints)
    assert "windspeed_120m" in env.state.dtype.names
    assert (env.state["windspeed_120m"] > env.state["windspeed_100m"]).all()

    env2 = Environment(state=weather, ws_alpha=0.12)

    assert "windspeed_120m" not in env2.state.dtype.names
    constraints = {"waveheight": le(2), "windspeed_120m": le(10)}
    valid = env2._find_valid_constraints(**constraints)
    assert (env.state["windspeed_100m"] == env2.state["windspeed_100m"]).all()
    assert (env.state["windspeed_120m"] < env2.state["windspeed_120m"]).all()



"""`ExportCableInstallation` and related processes."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy
from math import ceil

from marmot import process

from wisdem.orbit.core.logic import position_onsite
from wisdem.orbit.phases.install import InstallPhase
from wisdem.orbit.core.exceptions import InsufficientCable

from .common import SimpleCable as Cable
from .common import (
    lay_cable,
    bury_cable,
    dig_trench,
    pull_in_cable,
    landfall_tasks,
    lay_bury_cable,
    splice_process,
    terminate_cable,
    load_cable_on_vessel,
)


class ExportCableInstallation(InstallPhase):
    """Export Cable Installation Phase"""

    phase = "Export Cable Installation"
    capex_category = "Export System"

    #:
    expected_config = {
        "landfall": {"trench_length": "km (optional)"},
        "export_cable_install_vessel": "str | dict",
        "export_cable_bury_vessel": "str | dict (optional)",
        "export_cable_trench_vessel": "str (optional)",
        "site": {"distance": "km"},
        "plant": {"capacity": "MW"},
        "export_system": {
            "cable": {
                "linear_density": "t/km",
                "sections": [("length, km", "speed, km/h (optional)")],
                "number": "int (optional)",
            },
            "interconnection_distance": "km (optional); default: 3km",
            "interconnection_voltage": "kV (optional); default: 345kV",
        },
    }

    def __init__(self, config, weather=None, **kwargs):
        """
        Creates an instance of ExportCableInstallation.

        Parameters
        ----------
        config : dict
            Simulation specific configuration.
        weather : np.ndarray
            Weather profile at site.
        """

        super().__init__(weather, **kwargs)

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.initialize_port()
        self.extract_distances()
        self.setup_simulation(**kwargs)

    def setup_simulation(self, **kwargs):
        """
        Setup method for the ExportCableInstallation phase.
        - Extracts key inputs
        - Performs onshore infrastructure construction
        - Routes to specific setup scripts based on configured strategy.
        """

        depth = self.config["site"]["depth"]
        system = self.config["export_system"]
        self.free_cable_length = system.get("free_cable_length", depth / 1000)

        self.cable = Cable(system["cable"]["linear_density"])
        self.sections = system["cable"]["sections"]
        self.number = system["cable"].get("number", 1)

        self.initialize_installation_vessel()
        self.initialize_burial_vessel()
        self.initialize_trench_vessel()

        # Perform onshore construction
        onshore = kwargs.get("include_onshore_construction", True)
        if onshore:
            self.onshore_construction(**kwargs)

        # Perform cable installation
        install_export_cables(
            self.install_vessel,
            sections=self.sections,
            cable=self.cable,
            number=self.number,
            distances=self.distances,
            burial_vessel=self.bury_vessel,
            trench_vessel=self.trench_vessel,
            free_cable_length=self.free_cable_length,
            **kwargs,
        )

    @property
    def system_capex(self):
        """Returns total procurement cost of the array system."""

        return self.config["export_system"]["system_cost"]

    def extract_distances(self):
        """Extracts distances from input configuration or default values."""

        site = self.config["site"]["distance"]
        try:
            trench = self.config["landfall"]["trench_length"]

        except KeyError:
            trench = 1

        self.distances = {"site": site, "trench": trench}

    def onshore_construction(self, **kwargs):
        """
        Performs onshore construction prior to the installation of the export
        cable system.

        Parameters
        ----------
        construction_time : int | float
            Amount of time onshore construction takes.
            Default: 48h
        construction_rate : int | float
            Day rate of onshore construction.
            Default: 50000 USD/day
        """

        construction_time = kwargs.get("onshore_construction_time", 0.0)
        construction_cost = self.calculate_onshore_transmission_cost(**kwargs)

        if construction_time:
            _ = self.env.timeout(construction_time)
            self.env.run()

        self.env._submit_log(
            {
                "action": "Onshore Construction",
                "agent": "Onshore Construction",
                "duration": construction_time,
                "cost": construction_cost,
                "location": "Landfall",
            },
            level="ACTION",
        )

    def calculate_onshore_transmission_cost(self, **kwargs):
        """
        Calculates the cost of onshore transmission costs. From legacy
        OffshoreBOS model.
        """

        capacity = self.config["plant"]["capacity"]

        voltage = self.config["export_system"].get("interconnection_voltage", 345)
        distance = self.config["export_system"].get("interconnection_distance", 3)

        switchyard_cost = 18115 * voltage + 165944
        onshore_substation_cost = (0.165 * 1e6) * capacity  # From BNEF Tomorrow's Cost of Offshore Wind
        onshore_misc_cost = 11795 * capacity**0.3549 + 350000
        transmission_line_cost = (1176 * voltage + 218257) * (distance ** (1 - 0.1063))

        onshore_transmission_cost = (
            switchyard_cost + onshore_substation_cost + onshore_misc_cost + transmission_line_cost
        )

        return onshore_transmission_cost

    def initialize_installation_vessel(self):
        """Creates the export cable installation vessel."""

        # Vessel name and costs
        vessel_specs = self.config.get("export_cable_install_vessel", None)
        name = vessel_specs.get("name", "Export Cable Installation Vessel")

        vessel = self.initialize_vessel(name, vessel_specs)
        self.env.register(vessel)

        vessel.initialize()
        self.install_vessel = vessel

    def initialize_burial_vessel(self):
        """Creates the export cable burial vessel."""

        # Vessel name and costs
        vessel_specs = self.config.get("export_cable_bury_vessel", None)
        if vessel_specs is None:
            self.bury_vessel = None
            return

        name = vessel_specs.get("name", "Export Cable Burial Vessel")

        vessel = self.initialize_vessel(name, vessel_specs)
        self.env.register(vessel)

        vessel.initialize()
        self.bury_vessel = vessel

    def initialize_trench_vessel(self):
        """Creates the export cable trenching vessel."""

        # Vessel name and costs
        vessel_specs = self.config.get("export_cable_trench_vessel", None)
        if vessel_specs is None:
            self.trench_vessel = None
            return
        name = vessel_specs.get("name", "Export Cable Trench Vessel")

        vessel = self.initialize_vessel(name, vessel_specs)
        self.env.register(vessel)

        vessel.initialize()
        vessel.at_port = True
        vessel.at_site = False
        self.trench_vessel = vessel

    @property
    def detailed_output(self):
        """Detailed outputs of the export system installation."""

        outputs = {self.phase: {**self.agent_efficiencies}}

        return outputs


@process
def install_export_cables(
    vessel,
    sections,
    cable,
    number,
    distances,
    burial_vessel=None,
    trench_vessel=None,
    free_cable_length=None,
    **kwargs,
):
    """
    Simulation of the installation of export cables.

    Parameters
    ----------
    vessel : Vessel
        Cable installation vessel.
    sections : float
        Section lengths of the export cable.
    cable : SimpleCable | Cable
        Cable type to use.
    number : int
        Number of export cables.
    distances : dict
        Distances required for export cable installation simulation:
        site : int | float
            Distance between from the offshore substation and port. For
            simplicity, the cable landfall point is assumed to be at port.
        trench : int | float
            Trench length at landfall. Determines time required to tow the plow
            and pull-in cable (km).
    burial_vessel : Vessel
        Optional configuration for burial vessel. If configured, the
        installation vessel only lays the cable on the seafloor and this
        vessel will bury them at the end of the simulation.
    trench_vessel: Vessel
        Optional configuration for trenching vessel.  If configured, the
        trenching vessel travels along the cable route prior to arrival of
        the cable lay vessel and digs a trench.
    """

    ground_distance = -free_cable_length
    for s in sections:
        try:
            length, speed = s

        except TypeError:
            length = s

        ground_distance += length

    # Conduct trenching operations
    if trench_vessel is None:
        pass

    else:
        for _ in range(number):
            # Trenching vessel can dig a trench during inbound or outbound journey
            if trench_vessel.at_port:
                trench_vessel.at_port = False
                yield dig_export_cables_trench(trench_vessel, ground_distance, **kwargs)
                trench_vessel.at_site = True
            elif trench_vessel.at_site:
                trench_vessel.at_site = False
                yield dig_export_cables_trench(trench_vessel, ground_distance, **kwargs)
                trench_vessel.at_port = True

        # If the vessel finishes trenching at site, return to shore
        # TODO: replace with demobilization method
        if trench_vessel.at_site:
            trench_vessel.at_site = False
            yield trench_vessel.transit(ground_distance, **kwargs)
        trench_vessel.at_port = True

    for _ in range(number):
        vessel.cable_storage.reset()
        yield load_cable_on_vessel(vessel, cable, **kwargs)

        # At Landfall
        yield landfall_tasks(vessel, distances["trench"], **kwargs)

        for s in sections:
            splice_required = False
            try:
                length, speed = s
                if burial_vessel is None:
                    specs = {**kwargs, "cable_lay_bury_speed": speed}

                else:
                    specs = {**kwargs, "cable_lay_speed": speed}

            except TypeError:
                length = s
                specs = deepcopy(kwargs)

            remaining = length
            while remaining > 0:
                if splice_required:
                    yield splice_process(vessel, **kwargs)

                try:
                    section = vessel.cable_storage.get_cable(remaining)

                except InsufficientCable as e:
                    section = vessel.cable_storage.get_cable(e.current)

                if burial_vessel is None:
                    yield lay_bury_cable(vessel, section, **specs)

                else:
                    yield lay_cable(vessel, section, **specs)

                remaining -= ceil(section)
                if remaining > 0:
                    splice_required = True

                    yield vessel.transit(distances["site"])
                    vessel.cable_storage.reset()
                    yield load_cable_on_vessel(vessel, cable, **kwargs)
                    yield vessel.transit(distances["site"])

        # At Site
        yield position_onsite(vessel, **kwargs)
        yield pull_in_cable(vessel, **kwargs)
        yield terminate_cable(vessel, **kwargs)

        # Transit back to port
        yield vessel.transit(distances["site"])

    if burial_vessel is None:
        vessel.submit_debug_log(
            message="Export cable lay/burial process completed!",
            progress="Export System",
        )

    else:
        vessel.submit_debug_log(message="Export cable lay process completed!")
        bury_export_cables(burial_vessel, ground_distance, number, **kwargs)


@process
def bury_export_cables(vessel, length, number, **kwargs):
    """
    Simulation for the burial of export cables if configured.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel.
    length : float
        Full length of an export cable.
    number : int
        Number of export cables.
    """

    for _ in range(number):
        yield bury_cable(vessel, length, **kwargs)

    vessel.submit_debug_log(
        message="Export cable burial process completed!",
        progress="Export System",
    )


@process
def dig_export_cables_trench(vessel, distance, **kwargs):
    """
    Simulation for digging a trench for the export cables (if configured).

    Parameters
    ----------
    vessel : Vessel
        Performing vessel.
    distance : int | float
        Distance along export cable route to dig trench for cable
    """

    yield position_onsite(vessel, site_position_time=2)
    yield dig_trench(vessel, distance, **kwargs)

    vessel.submit_debug_log(message="Export cable trench digging process completed!")



#!/usr/bin/env python
# encoding: utf-8

# setup.py
# only if building in place: ``python setup.py build_ext --inplace``
import os
import re
import shutil
import platform
import subprocess
import setuptools

#######
# This forces wheels to be platform specific
from setuptools.dist import Distribution
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

class bdist_wheel(_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        self.root_is_pure = False

class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""
    def has_ext_modules(foo):
        return True
#######
    
def run_meson_build(staging_dir):
    prefix = os.path.join(os.getcwd(), staging_dir)
    purelibdir = "."

    # check if meson extra args are specified
    meson_args = ""
    if "MESON_ARGS" in os.environ:
        meson_args = os.environ["MESON_ARGS"]

    if platform.system() == "Windows":
        if not "FC" in os.environ:
            os.environ["FC"] = "gfortran"
        if not "CC" in os.environ:
            os.environ["CC"] = "gcc"

    # configure
    meson_path = shutil.which("meson")
    if meson_path is None:
        raise OSError("The meson command cannot be found on the system")

    meson_call = [meson_path, "setup", staging_dir, "--wipe",
                  f"--prefix={prefix}", f"-Dpython.purelibdir={purelibdir}",
                  f"-Dpython.platlibdir={purelibdir}"] + meson_args.split()
    meson_call = [m for m in meson_call if m.strip() != ""]
    print(meson_call)
    p1 = subprocess.run(meson_call, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    os.makedirs(staging_dir, exist_ok=True)
    setup_log = os.path.join(staging_dir, "setup.log")
    with open(setup_log, "wb") as f:
        f.write(p1.stdout)
    if p1.returncode != 0:
        with open(setup_log, "r") as f:
            print(f.read())
        raise OSError(meson_call, f"The meson setup command failed! Check the log at {setup_log} for more information.")

    # build
    meson_call = [meson_path, "compile", "-vC", staging_dir]
    meson_call = [m for m in meson_call if m != ""]
    print(meson_call)
    p2 = subprocess.run(meson_call, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    compile_log = os.path.join(staging_dir, "compile.log")
    with open(compile_log, "wb") as f:
        f.write(p2.stdout)
    if p2.returncode != 0:
        with open(compile_log, "r") as f:
            print(f.read())
        raise OSError(meson_call, f"The meson compile command failed! Check the log at {compile_log} for more information.")


def copy_shared_libraries():
    build_path = os.path.join(staging_dir, "wisdem")
    for root, _dirs, files in os.walk(build_path):
        for file in files:
            # move wisdem libraries to just under staging_dir
            if file.endswith((".so", ".lib", ".pyd", ".pdb", ".dylib", ".dll")):
                if ".so.p" in root or ".pyd.p" in root:  # excludes intermediate object files
                    continue
                file_path = os.path.join(root, file)
                new_path = str(file_path)
                match = re.search(staging_dir, new_path)
                new_path = new_path[match.span()[1] + 1 :]
                print(f"Copying build file {file_path} -> {new_path}")
                shutil.copy(file_path, new_path)


if __name__ == "__main__":
    # This is where the meson build system will install to, it is then
    # used as the sources for setuptools
    staging_dir = "meson_build"

    # this keeps the meson build system from running more than once
    if "dist" not in str(os.path.abspath(__file__)):
        cwd = os.getcwd()
        run_meson_build(staging_dir)
        os.chdir(cwd)
        copy_shared_libraries()

    # docs_require = ""
    # req_txt = os.path.join("doc", "requirements.txt")
    # if os.path.isfile(req_txt):
    #    with open(req_txt) as f:
    #        docs_require = f.read().splitlines()

    #init_file = os.path.join("wisdem", "__init__.py")
    # __version__ = re.findall(
    #    r"""__version__ = ["']+([0-9\.]*)["']+""",
    #    open(init_file).read(),
    # )[0]

    setuptools.setup(cmdclass={'bdist_wheel': bdist_wheel}, distclass=BinaryDistribution)

# os.environ['NPY_DISTUTILS_APPEND_FLAGS'] = '1'

# bemExt = Extension(
#    "wisdem.ccblade._bem",
#    sources=[os.path.join("wisdem", "ccblade", "src", "bem.f90")],
#    extra_compile_args=["-O2", "-fPIC", "-std=c11"],
# )
# pyframeExt = Extension(
#    "wisdem.pyframe3dd._pyframe3dd",
#    sources=glob.glob(os.path.join("wisdem", "pyframe3dd", "src", "*.c")),
#    extra_compile_args=["-O2", "-fPIC", "-std=c11"],
# )
# precompExt = Extension(
#    "wisdem.rotorse._precomp",
#    sources=[os.path.join("wisdem", "rotorse", "PreCompPy.f90")],
#    extra_compile_args=["-O2", "-fPIC", "-std=c11"],
# )

# Top-level setup
#    ext_modules=[bemExt, pyframeExt, precompExt],



#!/usr/bin/env python3

# Tower analysis
# Optimization by flag
# Two load cases
import os

import numpy as np
import openmdao.api as om

from wisdem.towerse.tower import TowerSE
from wisdem.commonse.fileIO import save_data

# Set analysis and optimization options and define geometry
plot_flag = False
opt_flag = True

n_control_points = 3
n_materials = 1
n_load_cases = 2

h_param = np.diff(np.linspace(0.0, 87.6, n_control_points))
d_param = np.linspace(8.0, 3.87, n_control_points)
t_param = np.linspace(0.08, 0.02, n_control_points)
max_diam = 9.0
# ---

# Store analysis options in dictionary
modeling_options = {}
modeling_options["flags"] = {}
modeling_options["materials"] = {}
modeling_options["materials"]["n_mat"] = n_materials
modeling_options["General"] = {}
modeling_options["WISDEM"] = {}
modeling_options["WISDEM"]["n_dlc"] = n_load_cases
modeling_options["WISDEM"]["TowerSE"] = {}
modeling_options["WISDEM"]["TowerSE"]["buckling_method"] = "eurocode"
modeling_options["WISDEM"]["TowerSE"]["buckling_length"] = 15.0
modeling_options["WISDEM"]["TowerSE"]["n_refine"] = 3

# safety factors
modeling_options["WISDEM"]["TowerSE"]["gamma_f"] = 1.35
modeling_options["WISDEM"]["TowerSE"]["gamma_m"] = 1.3
modeling_options["WISDEM"]["TowerSE"]["gamma_n"] = 1.0
modeling_options["WISDEM"]["TowerSE"]["gamma_b"] = 1.1
modeling_options["WISDEM"]["TowerSE"]["gamma_fatigue"] = 1.35 * 1.3 * 1.0

# Frame3DD options
modeling_options["WISDEM"]["TowerSE"]["frame3dd"] = {}
modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["shear"] = True
modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["geom"] = True
modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["tol"] = 1e-9
modeling_options["WISDEM"]["TowerSE"]["frame3dd"]["modal_method"] = 1
modeling_options["WISDEM"]["TowerSE"]["rank_and_file"] = True

modeling_options["WISDEM"]["TowerSE"]["n_height"] = n_control_points
modeling_options["WISDEM"]["TowerSE"]["n_layers"] = 1
modeling_options["WISDEM"]["TowerSE"]["wind"] = "PowerWind"
# ---

# Instantiate OpenMDAO problem and create a model using the TowerSE group
prob = om.Problem(reports=False)
prob.model = TowerSE(modeling_options=modeling_options)
# ---

# If performing optimization, set up the optimizer and problem formulation
if opt_flag:
    # Choose the optimizer to use
    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options["optimizer"] = "SLSQP"

    # Add objective
    prob.model.add_objective("tower_mass", scaler=1e-6)

    # Add design variables, in this case the tower diameter and wall thicknesses
    prob.model.add_design_var("tower_outer_diameter_in", lower=3.87, upper=max_diam)
    prob.model.add_design_var("tower_layer_thickness", lower=4e-3, upper=2e-1)

    # Add constraints on the tower design
    prob.model.add_constraint("post.constr_stress", upper=1.0)
    prob.model.add_constraint("post.constr_global_buckling", upper=1.0)
    prob.model.add_constraint("post.constr_shell_buckling", upper=1.0)
    prob.model.add_constraint("constr_d_to_t", lower=80.0, upper=500.0)
    prob.model.add_constraint("constr_taper", lower=0.2)
    prob.model.add_constraint("slope", upper=1.0)
    prob.model.add_constraint("tower.f1", lower=0.13, upper=0.40)
    # ---


# Set up the OpenMDAO problem
prob.setup()
# ---

# Set geometry and turbine values
prob["hub_height"] = prob["tower_height"] = h_param.sum()
prob["foundation_height"] = 0.0
prob["tower_s"] = np.cumsum(np.r_[0.0, h_param]) / h_param.sum()
prob["tower_outer_diameter_in"] = d_param
prob["tower_layer_thickness"] = t_param.reshape((1, -1))
prob["outfitting_factor_in"] = 1.07
prob["yaw"] = 0.0

# material properties
prob["E_mat"] = 210e9 * np.ones((n_materials, 3))
prob["G_mat"] = 80.8e9 * np.ones((n_materials, 3))
prob["rho_mat"] = [8500.0]
prob["sigma_y_mat"] = [450e6]
prob["sigma_ult_mat"] = 600e6 * np.ones((n_materials, 3))
prob["wohler_exp_mat"] = [10.0]
prob["wohler_A_mat"] = [10.0]

# extra mass from RNA
prob["rna_mass"] = np.array([285598.8])
mIxx = 1.14930678e08
mIyy = 2.20354030e07
mIzz = 1.87597425e07
mIxy = 0.0
mIxz = 5.03710467e05
mIyz = 0.0
prob["rna_I"] = np.array([mIxx, mIyy, mIzz, mIxy, mIxz, mIyz])
prob["rna_cg"] = np.array([-1.13197635, 0.0, 0.50875268])
# ---

# cost rates
prob["unit_cost_mat"] = [2.0]  # USD/kg
prob["labor_cost_rate"] = 100.0 / 60.0  # USD/min
prob["painting_cost_rate"] = 30.0  # USD/m^2

# wind & wave values
prob["wind_reference_height"] = 90.0
prob["z0"] = 0.0
prob["cd_usr"] = -1.0
prob["rho_air"] = 1.225
prob["mu_air"] = 1.7934e-5
prob["beta_wind"] = 0.0
if modeling_options["WISDEM"]["TowerSE"]["wind"] == "PowerWind":
    prob["shearExp"] = 0.2
# ---


# two load cases.  TODO: use a case iterator

# --- loading case 1: max Thrust ---
prob["env1.Uref"] = 11.73732
Fx1 = 1284744.19620519
Fy1 = 0.0
Fz1 = -2914124.84400512
Mxx1 = 3963732.76208099
Myy1 = -2275104.79420872
Mzz1 = -346781.68192839
# ---------------

# --- loading case 2: max Wind Speed ---
prob["env2.Uref"] = 70.0
Fx2 = 930198.60063279
Fy2 = 0.0
Fz2 = -2883106.12368949
Mxx2 = -1683669.22411597
Myy2 = -2522475.34625363
Mzz2 = 147301.97023764
# ---------------
prob["tower.rna_F"] = np.c_[[Fx1, Fy1, Fz1], [Fx2, Fy2, Fz2]]
prob["tower.rna_M"] = np.c_[[Mxx1, Myy1, Mzz1], [Mxx2, Myy2, Mzz2]]

# run the analysis or optimization
prob.model.approx_totals()
if opt_flag:
    prob.run_driver()
else:
    prob.run_model()
os.makedirs("outputs", exist_ok=True)
save_data(os.path.join("outputs", "tower_example"), prob)
# ---

# print results from the analysis or optimization
z = 0.5 * (prob["z_full"][:-1] + prob["z_full"][1:])
print("zs =", prob["z_full"])
print("ds =", prob["outer_diameter_full"])
print("ts =", prob["t_full"])
print("mass (kg) =", prob["tower_mass"])
print("cg (m) =", prob["tower_center_of_mass"])
print("d:t constraint =", prob["constr_d_to_t"])
print("taper ratio constraint =", prob["constr_taper"])
print("\nwind: ", prob["env1.Uref"])
print("freq (Hz) =", prob["tower.structural_frequencies"])
print("Fore-aft mode shapes =", prob["tower.fore_aft_modes"])
print("Side-side mode shapes =", prob["tower.side_side_modes"])
print("top_deflection (m) =", prob["tower.top_deflection"])
print("Tower base forces (N) =", prob["tower.turbine_F"])
print("Tower base moments (Nm) =", prob["tower.turbine_M"])
print("stress1 =", prob["post.constr_stress"])
print("GL buckling =", prob["post.constr_global_buckling"])
print("Shell buckling =", prob["post.constr_shell_buckling"])

if plot_flag:
    import matplotlib.pyplot as plt

    stress = prob["post.constr_stress"]
    shellBuckle = prob["post.constr_shell_buckling"]
    globalBuckle = prob["post.constr_global_buckling"]

    plt.figure(figsize=(5.0, 3.5))
    plt.subplot2grid((3, 3), (0, 0), colspan=2, rowspan=3)
    plt.plot(stress[:, 0], z, label="stress 1")
    plt.plot(stress[:, 1], z, label="stress 2")
    plt.plot(shellBuckle[:, 0], z, label="shell buckling 1")
    plt.plot(shellBuckle[:, 1], z, label="shell buckling 2")
    plt.plot(globalBuckle[:, 0], z, label="global buckling 1")
    plt.plot(globalBuckle[:, 1], z, label="global buckling 2")
    plt.legend(bbox_to_anchor=(1.05, 1.0), loc=2)
    plt.xlabel("utilization")
    plt.ylabel("height along tower (m)")
    plt.tight_layout()
    plt.show()
    # ---



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


from copy import deepcopy
from itertools import product

import pytest

from wisdem.orbit.phases.design import SemiSubmersibleDesign

base = {
    "site": {"depth": 500},
    "plant": {"num_turbines": 50},
    "turbine": {"turbine_rating": 12},
    "semisubmersible_design": {},
}


@pytest.mark.parametrize("depth,turbine_rating", product(range(100, 1201, 200), range(3, 15, 1)))
def test_parameter_sweep(depth, turbine_rating):
    config = {
        "site": {"depth": depth},
        "plant": {"num_turbines": 50},
        "turbine": {"turbine_rating": turbine_rating},
        "substation_design": {},
    }

    s = SemiSubmersibleDesign(config)
    s.run()

    assert s.detailed_output["stiffened_column_mass"] > 0
    assert s.detailed_output["truss_mass"] > 0
    assert s.detailed_output["heave_plate_mass"] > 0
    assert s.detailed_output["secondary_steel_mass"] > 0


def test_design_kwargs():
    test_kwargs = {
        "stiffened_column_CR": 3000,
        "truss_CR": 6000,
        "heave_plate_CR": 6000,
        "secondary_steel_CR": 7000,
    }

    s = SemiSubmersibleDesign(base)
    s.run()
    base_cost = s.total_cost

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["semisubmersible_design"] = {}
        config["semisubmersible_design"][k] = v

        s = SemiSubmersibleDesign(config)
        s.run()
        cost = s.total_cost

        assert cost != base_cost



import os
import glob
import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

from wisdem.commonse import fileIO


def clear_files():
    flist = glob.glob("test.*")
    for f in flist:
        os.remove(f)


class MyComp(om.ExplicitComponent):
    def setup(self):
        self.add_input("float_in", 0.0, units="N")
        self.add_input("fraction_in", 0.0)
        self.add_input("array_in", np.zeros(3), units="m")
        self.add_discrete_input("int_in", 0)
        self.add_discrete_input("string_in", "empty")
        self.add_discrete_input("list_in", ["empty"] * 3)

        self.add_output("float_out", 0.0, units="N")
        self.add_output("fraction_out", 0.0)
        self.add_output("array_out", np.zeros(3), units="m")
        self.add_discrete_output("int_out", 0)
        self.add_discrete_output("string_out", "empty")
        self.add_discrete_output("list_out", ["empty"] * 3)

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        outputs["float_out"] = inputs["float_in"] + 1
        outputs["fraction_out"] = inputs["fraction_in"] + 0.1
        outputs["array_out"] = inputs["array_in"] + 1
        discrete_outputs["int_out"] = 1
        discrete_outputs["string_out"] = "full"
        discrete_outputs["list_out"] = ["full"] * 3


class MyGroup(om.Group):
    def setup(self):
        self.add_subsystem("comp", MyComp(), promotes=["*"])


class TestFileIO(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem(reports=False, model=MyGroup())
        self.prob.setup()
        self.prob["float_in"] = 5.0
        self.prob.run_model()

    def tearDown(self):
        clear_files()

    def testSaveFile(self):
        clear_files()
        fileIO.save_data("test.junk", self.prob, mat_file=False, npz_file=False, xls_file=False)
        self.assertTrue(os.path.exists("test.pkl"))
        self.assertFalse(os.path.exists("test.npz"))
        self.assertFalse(os.path.exists("test.mat"))
        self.assertFalse(os.path.exists("test.xlsx"))

        clear_files()
        fileIO.save_data("test.junk", self.prob, mat_file=False)
        self.assertTrue(os.path.exists("test.pkl"))
        self.assertTrue(os.path.exists("test.npz"))
        self.assertFalse(os.path.exists("test.mat"))
        self.assertTrue(os.path.exists("test.xlsx"))

        clear_files()
        fileIO.save_data("test.junk", self.prob)
        self.assertTrue(os.path.exists("test.pkl"))
        self.assertTrue(os.path.exists("test.npz"))
        self.assertTrue(os.path.exists("test.mat"))
        self.assertTrue(os.path.exists("test.xlsx"))

    def testLoadFile(self):
        clear_files()
        fileIO.save_data("test", self.prob)
        self.prob = fileIO.load_data("test.pkl", self.prob)

        # Check pickle file
        self.assertEqual(self.prob["float_in"], 5.0)
        self.assertEqual(self.prob["float_out"], 6.0)
        self.assertEqual(self.prob["fraction_in"], 0.0)
        self.assertEqual(self.prob["fraction_out"], 0.1)
        npt.assert_equal(self.prob["array_in"], np.zeros(3))
        npt.assert_equal(self.prob["array_out"], np.ones(3))
        self.assertEqual(self.prob["int_in"], 0)
        self.assertEqual(self.prob["int_out"], 1)
        self.assertEqual(self.prob["string_in"], "empty")
        self.assertEqual(self.prob["string_out"], "full")
        self.assertEqual(self.prob["list_in"], ["empty"] * 3)
        self.assertEqual(self.prob["list_out"], ["full"] * 3)

        # Check numpy file
        npzdat = np.load("test.npz", allow_pickle=True)
        self.assertEqual(npzdat["float_in_N"], 5.0)
        self.assertEqual(npzdat["float_out_N"], 6.0)
        self.assertEqual(npzdat["fraction_in"], 0.0)
        self.assertEqual(npzdat["fraction_out"], 0.1)
        npt.assert_equal(npzdat["array_in_m"], np.zeros(3))
        npt.assert_equal(npzdat["array_out_m"], np.ones(3))
        self.assertEqual(npzdat["int_in"], 0)
        self.assertEqual(npzdat["int_out"], 1)
        self.assertEqual(npzdat["string_in"], "empty")
        self.assertEqual(npzdat["string_out"], "full")
        npt.assert_equal(npzdat["list_in"], ["empty"] * 3)
        npt.assert_equal(npzdat["list_out"], ["full"] * 3)


if __name__ == "__main__":
    unittest.main()



__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"

from .standard import ScourProtectionInstallation
